#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : objectives.py                                                     #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Tuesday, May 19th 2020, 1:06:30 pm                          #
# Last Modified : Tuesday, May 19th 2020, 1:06:31 pm                          #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Objective functions and their gradients."""
from abc import ABC, abstractmethod

import numpy as np
from sklearn.base import BaseEstimator

from mlstudio.utils.data_manager import GradientScaler    

# --------------------------------------------------------------------------  #
#                        OBJECTIVE BASE CLASS                                 #
# --------------------------------------------------------------------------  #
class Objective(ABC, BaseEstimator):
    """Base class for all objective functions.
    
    Parameters
    ----------
    regularizer : a Regularizer object
        Object to perform L1, L2, or L1_L2 regularization of the gradient

    gradient_scaler : a GradientScaler object or None
        Defaults to a GradientScaler object with default lower 
        threshold = 1e-10 and upper threshold = 1.
    """

    def __init__(self, regularizer=None, gradient_scaler=None):
        self.regularizer = regularizer
        self.gradient_scaler = gradient_scaler
        
    def _validation(self):
        from mlstudio.utils.validation import validate_regularizer                
        validate_regularizer(self.regularizer)
        if self.gradient_scaler:
            validate_gradient_scaler(self.gradient_scaler)

    def _check_gradient_scale(self, gradient):
        if self.gradient_scaler:
            gradient = self.gradient_scaler.fit_transform(gradient)
        return gradient

    @property
    def turn_off_gradient_scaling(self):
        self.gradient_scaler = False

    @abstractmethod
    def name(self):
        pass

    @abstractmethod
    def __call__(self, theta, **kwargs):
        """Computes the objective function.

        Parameters
        ----------
        theta : array-like
            The parameters from the model

        kwargs : Arbitrary keyword arguments.

        Returns
        -------
        real number
        """
        pass

    @abstractmethod
    def gradient(self, theta, **kwargs):
        """Computes the derivative of the objective function w.r.t. theta.
        
        Parameters
        ----------
        theta : array-like
            The parameters from the model

        kwargs : Arbitrary keyword arguments.

        Returns
        -------
        gradient : array-like in the shape of theta 
            The gradient of the objective function
        """
        pass
         

# --------------------------------------------------------------------------  #
class Cost(Objective):
    """Base class for all cost classes."""

    def __init__(self, regularizer=None, gradient_scaler=None):
        self.regularizer = regularizer
        self.gradient_scaler = gradient_scaler
    
    @property
    def name(self):
        return "Cost Base Class"        

    @abstractmethod
    def __call__(self, theta, y, y_out):
        pass

    @abstractmethod
    def gradient(self, theta, X, y, y_out):
        pass

# --------------------------------------------------------------------------  #
class MSE(Cost):

    @property
    def name(self):
        return "Mean Squared Error Cost Function"        

    def __call__(self, theta, y, y_out):
        """Computes the mean squared error cost.

        Parameters
        ----------
        y : array of shape (n_features,)
            Ground truth target values

        y_out : array of shape (n_features,)
            Output from the model 

        theta : array of shape (n_features,)  
            The model parameters            

        Returns
        -------
        cost : The quadratic cost 

        """   
        # Number of samples in the dataset     
        m = y.shape[0]     
        # Compute unregularized cost
        J = 0.5 * np.mean((y_out-y)**2)        
        # Apply regularization
        if self.regularizer:
            J += self.regularizer(theta, m)
        # Normalize cost and regularization by 
        return J

    def gradient(self, theta, X, y, y_out):
        """Computes quadratic costs gradient with respect to weights.
        
        Parameters
        ----------
        X : array of shape (m_observations, n_features)
            Input data

        y : array of shape (n_features,)
            Ground truth target values

        y_out : array of shape (n_features,)
            Output from the model 

        theta : array of shape (n_features,)  
            The model parameters                        

        Returns
        -------
        gradient of the cost function w.r.t. the parameters.

        """
        # Number of samples in the dataset
        m = y.shape[0]
        # Compute unregularized gradient
        gradient = (1/m) * X.T.dot(y_out-y)                         
        # Apply regularization to the weights (not bias) in gradient
        if self.regularizer:
            gradient += self.regularizer.gradient(theta, m)              
        # Check any vanishing or exploding gradients          
        gradient = self._check_gradient_scale(gradient)                            
        return gradient        

# --------------------------------------------------------------------------  #
class CrossEntropy(Cost):

    @property
    def name(self):
        return "Cross Entropy Cost Function"        

    def __call__(self, theta, y, y_out):
        """Computes cross entropy cost.

        Parameters
        ----------
        y : array of shape (n_features,)
            Ground truth target values

        y_out : array of shape (n_features,)
            Output from the model 

        theta : array of shape (n_features,)  
            The model parameters            

        Returns
        -------
        cost : The quadratic cost 

        """
        # Number of samples in the dataset
        m = y.shape[0]
        # Prevent division by zero
        y_out = np.clip(y_out, 1e-15, 1-1e-15)    
        # Compute unregularized cost
        J = -np.mean(y * np.log(y_out) + (1-y) * np.log(1-y_out)) 
        # Apply regularization to the weights (not bias) in gradient
        if self.regularizer: 
            J += self.regularizer(theta, m)
        # Return cost as average of cross entropy
        return J 

    def gradient(self, theta, X, y, y_out):
        """Computes cross entropy cost  gradient with respect to weights.
        
        Parameters
        ----------
        X : array of shape (m_observations, n_features)
            Input data

        y : array of shape (n_features,)
            Ground truth target values

        y_out : array of shape (n_features,)
            Output from the model 

        theta : array of shape (n_features,)  
            The model parameters                        

        Returns
        -------
        gradient of the cost function w.r.t. the parameters.

        """
        # Number of samples in the dataset
        m = X.shape[0]        
        # Compute unregularized gradient
        gradient = (1/m) * X.T.dot(y_out-y)         
        # Apply regularization to the weights (not bias) in gradient
        if self.regularizer:
            gradient += self.regularizer.gradient(theta, m)                
        # Check any vanishing or exploding gradients
        gradient = self._check_gradient_scale(gradient)            
        return gradient          

# --------------------------------------------------------------------------  #
class CategoricalCrossEntropy(Cost):


    @property
    def name(self):
        return "Categorical Cross Entropy Cost Function"        

    def __call__(self, theta, y, y_out):
        """Computes categorical cross entropy cost.

        Parameters
        ----------
        y : array of shape (n_features,)
            Ground truth target values

        y_out : array of shape (n_features,)
            Output from the model 

        theta : array of shape (n_features,)  
            The model parameters            

        Returns
        -------
        cost : The quadratic cost 

        """
        # Number of samples in the dataset
        m = y.shape[0]
        # Prevent division by zero
        y_out = np.clip(y_out, 1e-15, 1-1e-15)    
        # Compute unregularized cost
        J = -np.mean(np.sum(y * np.log(y_out), axis=1))
        # Add regularizer of weights 
        if self.regularizer:
            J += self.regularizer(theta, m) 
        # Return cost as average of cross-entropy
        return J

    def gradient(self, theta, X, y, y_out):
        """Computes gradient of cross-entropy cost with respect to weights.
        
        Parameters
        ----------
        X : array of shape (m_observations, n_features)
            Input data

        y : array of shape (n_features,)
            Ground truth target values

        y_out : array of shape (n_features,)
            Output from the model 

        theta : array of shape (n_features,)  
            The model parameters                        

        Returns
        -------
        gradient of the cost function w.r.t. the parameters.

        """
        gradient = {}
        # Number of samples in the dataset
        m = y.shape[0]
        # Compute unregularized gradient
        gradient = -(1/m) * (X.T.dot(y-y_out))
        # Add regularizer of weights 
        if self.regularizer:
            gradient += self.regularizer.gradient(theta, m)     
        # Check gradient scale before applying regularization
        gradient = self._check_gradient_scale(gradient)                    
        return gradient                  
# --------------------------------------------------------------------------  #
#                         BENCHMARK FUNCTIONS                                 #        
# --------------------------------------------------------------------------  #
class Benchmark(Objective):
    """Base class for objective functions."""

    def __init__(self, regularizer=None, gradient_scaler=GradientScaler()):
        self.regularizer = regularizer
        self.gradient_scaler = gradient_scaler

    @property
    def name(self):
        return "Benchmark Base Class"

    @property
    def density(self):
        """The linspace density per point of difference between min and max.""" 
        return 2

    @property
    def start(self):
        """Returns a good start point."""
        pass

    @property
    def minimum(self):
        """Returns a good start point."""
        pass    

    @property
    def range(self):
        """Returns x and y ranges for plotting."""
        pass
   
    
    @abstractmethod
    def __call__(self, theta, **kwargs):
        """Computes the objective function value"""
        pass
    
    @abstractmethod
    def gradient(self, theta, **kwargs):
        """Computes the gradient of the objective function."""
        pass
# --------------------------------------------------------------------------  #
class Adjiman(Benchmark):
    """Base class for objective functions.""" 

    @property
    def name(self):
        return "Adjiman Objective Function"    

    @property   
    def start(self):
        return np.array([-4,0])

    @property
    def minimum(self):
        return np.array([2,0.10578])

    @property
    def range(self):
        """Returns the x and y ranges for plotting."""
        x = {'min': -4, 'max': 4}
        y = {'min': -5, 'max': 5}
        return x, y
    
    def __call__(self, theta, **kwargs):
        """Computes the objective function value"""
        return np.cos(theta[0]) * np.sin(theta[1]) - (theta[0] / (theta[1]**2 + 1))

    def gradient(self, theta, **kwargs):
        """Computes the gradient of the objective function."""
        dfdx = -(1/(theta[1]**2+1))*((theta[1]**2+1)*np.sin(theta[0])*np.sin(theta[1])+1)
        dfdy = 2*theta[0]*theta[1] /(theta[1]**2+1)**2 + np.cos(theta[0])*np.cos(theta[1])
        df = np.array([dfdx, dfdy])
        # Check gradient scale 
        df = self._check_gradient_scale(df)                
        return df

# --------------------------------------------------------------------------  #
class BartelsConn(Benchmark):
    """Base class for objective functions."""    

    @property
    def name(self):
        return "Bartels Conn Objective Function"
    
    @property
    def start(self):
        return np.array([-500,0])

    @property
    def minimum(self):
        return np.array([0,0])        

    @property
    def range(self):
        """Returns the x and y ranges for plotting."""
        x = {'min': -500, 'max': 500}
        y = {'min': -500, 'max': 500}
        return x, y
    
    def __call__(self, theta, **kwargs):
        """Computes the objective function value"""
        a = abs(theta[0]**2 + theta[1]**2 + theta[0] * theta[1])
        b = abs(np.sin(theta[0]))
        c = abs(np.cos(theta[1]))
        return a + b + c

    def gradient(self, theta, **kwargs):
        """Computes the gradient of the objective function."""
        a = (2 * theta[0] + theta[1]) * np.sign(theta[0]**2 + theta[0] * theta[1] + theta[1]**2)
        b = np.cos(theta[0]) * np.sign(np.sin(theta[0]))
        dfdx = a + b
        a = (theta[0] + 2 * theta[1]) * np.sign(theta[0]**2 + theta[0] * theta[1] + theta[1]**2)
        b = np.sin(theta[1]) * np.sign(np.cos(theta[1]))
        dfdy = a - b
        # Package into gradient vector
        df = np.array([dfdx, dfdy])
        # Check gradient scale 
        df = self._check_gradient_scale(df)                
        return df        

# --------------------------------------------------------------------------  #
class Himmelblau(Benchmark):
    """Base class for objective functions."""   

    @property
    def name(self):
        return "Himmelblau Objective Function"     

    @property
    def start(self):
        return np.array([5,5])

    @property
    def minimum(self):
        return np.array([3,2])           

    @property
    def range(self):
        """Returns the x and y ranges for plotting."""
        x = {'min': -5, 'max': 5}
        y = {'min': -5, 'max': 5}
        return x, y    

    def __call__(self, theta, **kwargs):
        """Computes the objective function value"""
        return (theta[0]**2 + theta[1] - 11)**2 + (theta[0]+theta[1]**2-7)**2

    def gradient(self, theta, **kwargs):
        """Computes the gradient of the objective function."""
        dfdx = 4*theta[0]*(theta[0]**2+theta[1]-11)+2*theta[0]+2*theta[1]**2-14
        dfdy = 2*theta[0]**2 + 4*theta[1] * (theta[0]+theta[1]**2-7)+2*theta[1]-22
        # Package into gradient vector
        df = np.array([dfdx, dfdy])
        # Check gradient scale 
        df = self._check_gradient_scale(df)                
        return df        
        

# --------------------------------------------------------------------------  #
class Leon(Benchmark):
    """Base class for objective functions."""    

    @property
    def name(self):
        return "Leon Objective Function"    

    @property
    def start(self):
        return np.array([-1.5,1.5])

    @property
    def minimum(self):
        return np.array([1,1])  

    @property
    def range(self):
        """Returns the x and y ranges for plotting."""
        x = {'min': -1.5, 'max': 1.5}
        y = {'min': -2, 'max': 2}
        return x, y            

    def __call__(self, theta, **kwargs):
        """Computes the objective function value"""
        return 100 * (theta[1] - theta[0]**2)**2 + (1 - theta[0])**2

    def gradient(self, theta, **kwargs):
        """Computes the gradient of the objective function."""
        dfdx = -400*(-theta[0]**2+theta[1]) + 2*theta[0] - 2
        dfdy = -200*theta[0]**2 + 200 * theta[1]
        # Package into gradient vector
        df = np.array([dfdx, dfdy])
        # Check gradient scale 
        df = self._check_gradient_scale(df)                
        return df        
             

   

# --------------------------------------------------------------------------  #
class Rosenbrock(Benchmark):
    """Base class for objective functions."""

    @property
    def name(self):
        return "Rosenbrock Objective Function"    

    @property
    def start(self):
        return np.array([-5,10])

    @property
    def minimum(self):
        return np.array([1,1])  

    @property
    def range(self):
        """Returns the x and y ranges for plotting."""
        x = {'min': -10, 'max': 10}
        y = {'min': -10, 'max': 10}
        return x, y            

    
    def __call__(self, theta, **kwargs):
        """Computes the objective function value"""
        a = 1
        b = 100
        return b*(theta[1]-theta[0]**2)**2+(a-theta[0])**2

    def gradient(self, theta, **kwargs):
        """Computes the gradient of the objective function."""
        dfdx = -400*theta[0]*(-theta[0]**2+theta[1])+2*theta[0]-2
        dfdy = -200*theta[0]**2 + 200 * theta[1]
        # Package into gradient vector
        df = np.array([dfdx, dfdy])
        # Check gradient scale 
        df = self._check_gradient_scale(df)                
        return df        

# --------------------------------------------------------------------------  #
class Branin02(Benchmark):
    """Branin02 objective functions."""    

    @property
    def name(self):
        return "Branin02 Objective Function"    

    @property
    def start(self):
        return np.array([-5,-5])
        
    @property
    def minimum(self):
        return np.array([-3.2,12.53])           
    
    @property
    def range(self):
        """Returns the x and y ranges for plotting."""
        x = {'min': -5, 'max': 15}
        y = {'min': -5, 'max': 15}
        return x, y
    
    def __call__(self, theta, **kwargs):
        """Computes the objective function value"""        
        return (-1.275*theta[0]**2/np.pi**2 + 5*theta[0]/np.pi + theta[1]-6)**2\
            + (10-5/4*np.pi) * np.cos(theta[0]) * np.cos(theta[1]) \
                + np.log(theta[0]**2+theta[1]**2+1) + 10

    def gradient(self, theta, **kwargs):
        """Computes the gradient of the objective function."""
        # dfdx
        a = 1/(4*np.pi**4*(theta[0]**2+theta[1]**2+1))
        b = 8 * np.pi**4 * theta[0]
        c = 5*(theta[0] - 2 * np.pi)
        d = (theta[0]**2+theta[1]**2+1)
        e = (5 * theta[0]**2 - 20*np.pi*theta[0]+4*np.pi**2*(6-theta[1]))
        f = 5*np.pi**3 *  (-1+8*np.pi) * (theta[0]**2+theta[1]**2+1) \
            * np.sin(theta[0]) * np.cos(theta[1])
        dfdx = a * (b+c*d*e - f)
        # dfdy
        a = -(5*theta[0]**2/(2*(np.pi**2)))
        b = (10 * theta[0])/np.pi
        c = 2 * theta[1]
        d = (2*theta[1]) / (theta[0]**2+theta[1]**2+1)
        e = (10-(5/(np.pi*4)))
        f = np.sin(theta[1]) * np.cos(theta[0])
        dfdy = a + b + c + d - e * f - 12
        # Package into gradient vector
        df = np.array([dfdx, dfdy])
        # Check gradient scale 
        df = self._check_gradient_scale(df)                
        return df        


# --------------------------------------------------------------------------  #
class StyblinskiTank(Benchmark):
    """Styblinksi-Tank objective functions."""

    @property
    def name(self):
        return "Styblinski-Tank Objective Function"    

    @property
    def start(self):
        return np.array([-5,-4])

    @property
    def minimum(self):
        return np.array([-2.903534, -2.903534])  

    @property
    def range(self):
        """Returns the x and y ranges for plotting."""
        x = {'min': -5, 'max': 5}
        y = {'min': -5, 'max': 5}
        return x, y            

    
    def __call__(self, theta, **kwargs):
        """Computes the objective function value"""   
        a = 0
        for i in range(len(theta, **kwargs)):
            a += theta[i]**4 - 16 * theta[i]**2 + 5 * theta[i]

        return 1/2 * a

    def gradient(self, theta, **kwargs):
        """Computes the gradient of the objective function."""
        dfdx = 2*theta[0]**3 - 16 * theta[0] + 5/2
        dfdy = 2*theta[1]**3 - 16*theta[1] + 5/2
        # Package into gradient vector
        df = np.array([dfdx, dfdy])
        # Check gradient scale 
        df = self._check_gradient_scale(df)                
        return df        
       

# --------------------------------------------------------------------------  #
class SumSquares(Benchmark):
    """Base class for objective functions."""    

    @property
    def name(self):
        return "Sum Squares Objective Function"    

    @property
    def start(self):
        return np.array([10,10])
        
    @property
    def minimum(self):
        return np.array([0,0])           
    
    @property
    def range(self):
        """Returns the x and y ranges for plotting."""
        x = {'min': -10, 'max': 10}
        y = {'min': -10, 'max': 10}
        return x, y
    
    def __call__(self, theta, **kwargs):
        """Computes the objective function value"""        
        return np.sum(theta[0]**2 + theta[1]**2)

    def gradient(self, theta, **kwargs):
        """Computes the gradient of the objective function."""
        dfdx = 2 * theta[0]
        dfdy = 2 * theta[1]
        # Package into gradient vector
        df = np.array([dfdx, dfdy])
        # Check gradient scale 
        df = self._check_gradient_scale(df)                
        return df                

# --------------------------------------------------------------------------  #
class ThreeHumpCamel(Benchmark):
    """Three hump camel objective functions."""    

    @property
    def name(self):
        return "Three Hump Camel Objective Function"    

    @property
    def start(self):
        return np.array([-1.5,1.5])
        
    @property
    def minimum(self):
        return np.array([0,0])           
    
    @property
    def range(self):
        """Returns the x and y ranges for plotting."""
        x = {'min': -2, 'max': 2}
        y = {'min': -2, 'max': 2}
        return x, y
    
    def __call__(self, theta, **kwargs):
        """Computes the objective function value"""        
        return 2*theta[0]**2 - 1.05*theta[0]**4 + ((theta[0]**6)/6)+theta[0]*theta[1]+theta[1]**2

    def gradient(self, theta, **kwargs):
        """Computes the gradient of the objective function."""
        dfdx = theta[0]**5-((21*theta[0]**3)/5)+4*theta[0]+theta[1]
        dfdy = theta[0]+2*theta[1]
        # Package into gradient vector
        df = np.array([dfdx, dfdy])
        # Check gradient scale 
        df = self._check_gradient_scale(df)                
        return df        

# --------------------------------------------------------------------------  #
class Ursem01(Benchmark):
    """Ursem01 objective functions."""    

    @property
    def name(self):
        return "Ursem01 Objective Function"    

    @property
    def start(self):
        return np.array([0,2.5])
        
    @property
    def minimum(self):
        return np.array([1.69714,0])           
    
    @property
    def range(self):
        """Returns the x and y ranges for plotting."""
        x = {'min': -2, 'max': 3}
        y = {'min': -2, 'max': 2.5}
        return x, y
    
    def __call__(self, theta, **kwargs):
        """Computes the objective function value"""        
        return -np.sin(2*theta[0] - 0.5 * np.pi) - 3 * np.cos(theta[1]) - 0.5 * theta[0]

    def gradient(self, theta, **kwargs):
        """Computes the gradient of the objective function."""
        dfdx = -2 * np.sin(2 * theta[0]) - 0.5
        dfdy = 3 * np.sin(theta[1])
        # Package into gradient vector
        df = np.array([dfdx, dfdy])
        # Check gradient scale 
        df = self._check_gradient_scale(df)                
        return df        

# --------------------------------------------------------------------------  #
class Wikipedia(Benchmark):
    """Wikipedia objective functions."""    

    @property
    def name(self):
        return "Wikipedia Objective Function"    

    @property
    def start(self):
        return np.array([0,-1])
        
    @property
    def minimum(self):
        return np.array([-0.5,1])           
    
    @property
    def range(self):
        """Returns the x and y ranges for plotting."""
        x = {'min': -1, 'max': 1}
        y = {'min': -1, 'max': 1}
        return x, y
    
    def __call__(self, theta, **kwargs):
        """Computes the objective function value"""        
        return np.sin(1/2*theta[0]**2 - 0.25 * theta[1]**2 + 3) * np.cos(2*theta[0]+1-np.exp(theta[1]))

    def gradient(self, theta, **kwargs):
        """Computes the gradient of the objective function."""
        a = (theta[0] * np.cos(2*theta[0]-np.exp(theta[1])+1))
        b = np.cos((theta[0]**2)/2-(theta[1]**2)/4+3)
        c = -2*np.sin(2*theta[0])-np.exp(theta[1])+1
        d = np.sin((theta[0]**2)/2-(theta[1]**2)/4+3)
        e = -theta[1]/2*np.cos(2*theta[0]-np.exp(theta[1])+1)
        f = np.cos((theta[0]**2)/2-(theta[1]**2)/4+3)
        g = np.exp(theta[1])*np.sin(2*theta[0]-np.exp(theta[1])+1)
        h = np.sin((theta[0]**2)/2-(theta[1]**2)/4+3)
        dfdx = a * b + c * d
        dfdy = e * f + g * h
        # Package into gradient vector
        df = np.array([dfdx, dfdy])
        # Check gradient scale 
        df = self._check_gradient_scale(df)                
        return df        

          



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : test_py                                                #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Sunday, May 24th 2020, 5:15:40 am                           #
# Last Modified : Sunday, May 24th 2020, 5:15:40 am                           #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Tests validation utilities."""
import sys
import numpy as np
import pytest
from pytest import mark
import scipy.sparse as sp

from mlstudio.utils.validation import is_one_hot
from mlstudio.utils.validation import is_multilabel, is_valid_array_size
# --------------------------------------------------------------------------  #
@mark.utils
@mark.validation
@mark.data_checks
@mark.is_one_hot
def test_is_one_hot(get_data_management_data):
    d = get_data_management_data
    for k, y in d.items():        
        msg = "Is one-hot of " + k + " didn't work."
        if k == 'one_hot':
            assert is_one_hot(y), msg
        else:
            assert not is_one_hot(y), msg

# --------------------------------------------------------------------------  #
@mark.utils
@mark.validation
@mark.data_checks
@mark.is_multilabel
def test_is_multilabel(get_data_management_data):
    d = get_data_management_data
    for k, y in d.items():        
        msg = "Is multilabel of " + k + " didn't work."
        if 'multilabel' in k:
            assert is_multilabel(y), msg
        else:
            assert not is_multilabel(y), msg

# --------------------------------------------------------------------------  #
@mark.utils
@mark.validation
@mark.data_checks
def test_is_valid_array_size():
    from mlstudio.utils.validation import is_valid_array_size
    X = np.random.default_rng().uniform(low=100, high=200, size=20)       
    Y = np.random.default_rng().uniform(low=1, high=5, size=20)       
    assert not is_valid_array_size(X, lower=1, upper=2), "is_valid_array_size failed"
    assert is_valid_array_size(Y, lower=1, upper=100), "is_valid_array_size failed"
  
# --------------------------------------------------------------------------  #
@mark.utils
@mark.validation
def test_validate_zero_to_one():
    from mlstudio.utils.validation import validate_zero_to_one
    with pytest.raises(ValueError) as v:
        validate_zero_to_one(5, 'test_param')
        assert "assertion error"  in str(v.value)
    with pytest.raises(ValueError) as v:
        validate_zero_to_one(0, left='open')
        assert "assertion error"  in str(v.value)    
    with pytest.raises(ValueError) as v:
        validate_zero_to_one(1, right='open')
        assert "assertion error"  in str(v.value)    
    validate_zero_to_one(0, left="closed")


# --------------------------------------------------------------------------  #
@mark.utils
@mark.validation
def test_validate_range():
    from mlstudio.utils.validation import validate_range
    with pytest.raises(ValueError) as v:
        validate_range(param=1, minimum=0, maximum=1, param_name='test_param')
        assert "value error"  in str(v.value)
    with pytest.raises(ValueError) as v:
        validate_range(param=0, minimum=0, maximum=1, param_name='test_param')
        assert "value error"  in str(v.value)        
    validate_range(param=1, minimum=0, maximum=1, right='closed')
    validate_range(param=0, minimum=0, maximum=1, left='closed')

# --------------------------------------------------------------------------  #    

@mark.utils
@mark.validation
def test_validate_string():
    from mlstudio.utils.validation import validate_string
    valid_values = ['epoch', 'batch']
    with pytest.raises(ValueError) as v:
        validate_string('hand', valid_values=valid_values)
        assert "assertion error"  in str(v.value)
    validate_string('batch', valid_values=valid_values)

# --------------------------------------------------------------------------  #    
@mark.utils
@mark.validation
def test_validate_activation():    
    from mlstudio.utils.validation import validate_activation
    from mlstudio.supervised.core.activations import Sigmoid
    with pytest.raises(TypeError) as v:
        validate_activation('hand')
        assert "value error"  in str(v.value)
    validate_activation(Sigmoid())

# --------------------------------------------------------------------------  #    
@mark.utils
@mark.validation
def test_validate_objective():    
    from mlstudio.utils.validation import validate_objective
    from mlstudio.supervised.core.objectives import MSE
    with pytest.raises(TypeError) as v:
        validate_objective('hand')
        assert "value error"  in str(v.value)
    validate_objective(MSE())    

# --------------------------------------------------------------------------  #    
@mark.utils
@mark.validation
def test_validate_optimizer():    
    from mlstudio.utils.validation import validate_optimizer
    from mlstudio.supervised.core.optimizers import Adam 
    with pytest.raises(TypeError) as v:
        validate_optimizer('hand')
        assert "value error"  in str(v.value)
    validate_optimizer(Adam())     

# --------------------------------------------------------------------------  #    
@mark.utils
@mark.validation
def test_validate_regularizer():    
    from mlstudio.utils.validation import validate_regularizer
    from mlstudio.supervised.core.regularizers import L1
    with pytest.raises(TypeError) as v:
        validate_regularizer('hand')
        assert "value error"  in str(v.value)
    validate_regularizer(L1())       

# --------------------------------------------------------------------------  #    
@mark.utils
@mark.validation
def test_validate_scorer():    
    from mlstudio.utils.validation import validate_scorer
    from mlstudio.supervised.core.scorers import MSE    
    with pytest.raises(TypeError) as v:
        validate_scorer('hand')
        assert "value error"  in str(v.value)
    validate_scorer(MSE())           

# --------------------------------------------------------------------------  #    
@mark.utils
@mark.validation
def test_validate_task():    
    from mlstudio.utils.validation import validate_task
    from mlstudio.supervised.core.tasks import LogisticRegression    
    with pytest.raises(TypeError) as v:
        validate_task('hand')
        assert "value error"  in str(v.value)
    validate_task(LogisticRegression())           




#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.14                                                            #
# File    : test_regressor.py                                                 #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Friday, June 19th 2020, 4:29:58 am                          #
# Last Modified : Friday, June 19th 2020, 4:29:58 am                          #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Integration test for GradientDescentRegressor class."""
import math
import numpy as np
import pytest
from pytest import mark
from sklearn.linear_model import LinearRegression
from sklearn.utils.estimator_checks import parametrize_with_checks
from sklearn.utils.estimator_checks import check_estimator

from mlstudio.supervised.machine_learning.gradient_descent import GradientDescentRegressor
from mlstudio.supervised.observers.learning_rate import TimeDecay, StepDecay
from mlstudio.supervised.observers.learning_rate import ExponentialDecay
from mlstudio.supervised.observers.learning_rate import ExponentialStepDecay
from mlstudio.supervised.observers.learning_rate import PolynomialDecay
from mlstudio.supervised.observers.learning_rate import PolynomialStepDecay
from mlstudio.supervised.observers.learning_rate import PowerSchedule
from mlstudio.supervised.observers.learning_rate import BottouSchedule
from mlstudio.supervised.observers.learning_rate import Improvement
from mlstudio.supervised.observers.early_stop import Performance
from mlstudio.supervised.observers.debugging import GradientCheck
from mlstudio.supervised.core.objectives import MSE
from mlstudio.supervised.core.optimizers import GradientDescentOptimizer
from mlstudio.supervised.core.optimizers import Momentum
from mlstudio.supervised.core.optimizers import Nesterov
from mlstudio.supervised.core.optimizers import Adagrad
from mlstudio.supervised.core.optimizers import Adadelta
from mlstudio.supervised.core.optimizers import RMSprop
from mlstudio.supervised.core.optimizers import Adam, AdaMax, Nadam
from mlstudio.supervised.core.optimizers import AMSGrad, AdamW, QHAdam
from mlstudio.supervised.core.optimizers import QuasiHyperbolicMomentum
from mlstudio.supervised.core.regularizers import L1, L2, L1_L2
from mlstudio.supervised.core import scorers
# --------------------------------------------------------------------------  #
count = 0
observers = [[Performance(mode='passive')], [Performance(mode='active')],
            [TimeDecay()], [StepDecay()], [ExponentialDecay()], 
            [ExponentialStepDecay()], [PolynomialDecay()], [PolynomialStepDecay()], 
            [PowerSchedule()], [BottouSchedule()], [Improvement()]]
scorer_objects = [scorers.R2(), scorers.MSE()]
objectives = [MSE(), MSE(regularizer=L1(alpha=0.01)), 
                        MSE(regularizer=L2(alpha=0.01)), 
                        MSE(regularizer=L1_L2(alpha=0.01, ratio=0.5))]
# optimizers = [
#     GradientDescentOptimizer(), Momentum(), Nesterov(),Adagrad(), Adadelta(),
#     RMSprop(), Adam(), AdaMax(), Nadam(), AMSGrad(), AdamW(), QHAdam(),
#     QuasiHyperbolicMomentum()
# ]

scenarios = [[observer, scorer, objective] for observer in observers
                                           for scorer in scorer_objects
                                           for objective in objectives
        ]

estimators = [GradientDescentRegressor(observers=scenario[0], scorer=scenario[1],
                                       objective=scenario[2]) for scenario in scenarios]
@mark.gd
@mark.regressor_skl
@parametrize_with_checks(estimators)
def test_regression_sklearn(estimator, check):    
    observer = [o.name for o in estimator.observers]    
    #learning_rate = estimator.learning_rate.name
    objective = estimator.objective.name
    regularizer = estimator.objective.regularizer.name if estimator.objective.regularizer else\
        None
    # optimizer = estimator.optimizer.name
    msg = "Checking scenario : observers : {o}, objective : {ob},\
            regularizer : {r}".format(
                o=str(observer), ob=str(objective), r=str(regularizer))
    print(msg)        
    check(estimator)

@mark.gd
@mark.regressor
def test_regressor(get_regression_data_split):
    X_train, X_test, y_train, y_test = get_regression_data_split
    for estimator in estimators:
        # Extract scenario options
        try:
            observer = [o.name for o in estimator.observers]
        except:
            observer = [estimator.observers.name]
#        learning_rate = estimator.learning_rate.name
        objective = estimator.objective.name
        regularizer = estimator.objective.regularizer.name if estimator.objective.regularizer else\
            None        
        msg = "Checking scenario: observers : {o}, objective : {ob},\
            regularizer : {r}".format(o=str(observer), 
            ob=str(objective), r=str(regularizer))
        print(msg)                     
        # Fit the model
        estimator.fit(X_train,y_train)
        mls_score = estimator.score(X_test, y_test)
        # Fit sklearn's model
        skl = LinearRegression()
        skl.fit(X_train, y_train)
        skl_score = skl.score(X_test, y_test)
        if mls_score < skl_score:
            print(msg)
        msg = "Score is significantly worst than sklearn's score. MLS score = {m}, \
            Sklearn score = {s}".format(m=str(mls_score), s=str(skl_score))
        # assert np.isclose(mls_score, skl_score, rtol=0.01) or\
        #     mls_score > skl_score, msg
        estimator.summary()






#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : test_data_analysis.py                                             #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Monday, May 25th 2020, 4:37:35 am                           #
# Last Modified : Monday, May 25th 2020, 4:37:36 am                           #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Tests for data analysis functions."""
import numpy as np
import pytest
from pytest import mark
import string

from mlstudio.utils.data_analyzer import describe_categorical_array
from mlstudio.utils.data_analyzer import describe_numeric_array
# --------------------------------------------------------------------------  #
#                             TEST DESCRIBE                                   #
# --------------------------------------------------------------------------  #
@mark.utils
@mark.data_analysis
@mark.describe
def test_describe_numeric_array():
    a = np.random.randint(low=1, high=100, size=100)
    d = describe_numeric_array(a)
    columns = ['min', 'max', 'mean', 'std', 'skew', 'kurtosis', '25th percentile',
               '50th percentile', '75th percentile', 'count']
    assert len(d) == 10, "Describe numeric array dictionary wrong length."
    for c in columns:        
        assert d.get(c), "Describe numeric didn't return " + c

@mark.utils
@mark.data_analysis
@mark.describe
def test_describe_categorical_array():
    s = string.ascii_lowercase
    s = list(s)
    a = np.random.choice(s, size=100)
    d = describe_categorical_array(a)
    columns = ['count', 'top', 'unique', 'freq']
    assert len(d) == 4, "Describe categorical array dictionary wrong length."
    for c in columns:        
        assert d.get(c) is not None, "Describe categorical didn't return " + c







#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : test_data_management.py                                           #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Monday, May 11th 2020, 8:33:38 pm                           #
# Last Modified : Monday, May 11th 2020, 8:33:38 pm                           #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Tests data management utilities."""
#%%
import numpy as np
import pytest
from pytest import mark
from scipy.sparse import csr_matrix
from sklearn.datasets import make_classification

from mlstudio.utils.data_manager import MinMaxScaler, data_split, GradientScaler
from mlstudio.utils.data_manager import encode_labels
from mlstudio.utils.data_manager import AddBiasTerm, ZeroBiasTerm, unpack_parameters
# --------------------------------------------------------------------------  #
#                       TEST ADD BIAS TERM TRANSFORMER                        #
# --------------------------------------------------------------------------  #
@mark.utils
@mark.data_manager
@mark.add_bias_term
def test_add_bias_term_np():
    X = np.random.rand(5,5)
    xformer = AddBiasTerm()
    X = xformer.fit_transform(X)
    assert X.shape[1] == 6, "Bias term not added."
    assert np.all(X[:,0] == 1.0), "Column zero not ones."
    # Inverse transform
    X = xformer.inverse_transform(X)
    assert X.shape[1] == 5, "Bias term not removed."
    

@mark.utils
@mark.data_manager
@mark.add_bias_term
def test_add_bias_term_csr():
    X = np.random.rand(5,5)
    X = csr_matrix(X)
    xformer = AddBiasTerm()
    X = xformer.fit_transform(X)
    assert X.shape[1] == 6, "Bias term not added."
    assert np.all(X.toarray()[:,0] == 1.0), "Column zero not ones."   
    # Inverse transform
    X = xformer.inverse_transform(X)
    assert X.shape[1] == 5, "Bias term not removed."    

# --------------------------------------------------------------------------  #
#                       TEST ZERO BIAS TERM TRANSFORMER                       #
# --------------------------------------------------------------------------  #
@mark.utils
@mark.data_manager
@mark.zero_bias_term
def test_zero_bias_term():
    X = np.random.rand(5)
    xformer = ZeroBiasTerm()
    X = xformer.fit_transform(X)    
    assert X[0] == 0.0, "Bias not zeroed out ."
    X = np.random.rand(5,5)
    xformer = ZeroBiasTerm()
    X = xformer.fit_transform(X)    
    assert np.array_equal(X[0,:], np.zeros(shape=X.shape[1])), "Bias not zeroed out ."


# --------------------------------------------------------------------------  #
#                       TEST DATA CHECKS AND PREP                             #
# --------------------------------------------------------------------------  #
@mark.utils
@mark.data_manager
@mark.data_checks
@mark.encode_labels
def test_encode_labels(get_data_management_data):
    d = get_data_management_data
    for k, y in d.items():
        classes = np.unique(y)
        n_classes = len(classes)
        encoded_classes = np.arange(n_classes)
        y_new = encode_labels(y)
        y_new_classes = np.sort(np.unique(y_new))
        msg = "Encoding of " + k + " didn't work."
        assert np.array_equal(encoded_classes, y_new_classes), msg
  

# --------------------------------------------------------------------------  #
#                        TEST GRADIENT SCALER                                 #
# --------------------------------------------------------------------------  #  
@mark.utils
@mark.data_manager
@mark.gradient_scaler
@mark.gradient_scaler_1d
def test_gradient_scaler_1d():            
    lower_threshold = 1e-10
    upper_threshold = 1e10
    lows = [1e-20, 1e15, 1] 
    highs = [1e-10, 1e20, 5]
    for g in zip(lows, highs):    
        X = np.random.default_rng().uniform(low=g[0], high=g[1], size=20)                
        X_orig_norm = np.linalg.norm(X)        
        scaler = GradientScaler(lower_threshold=lower_threshold, 
                                upper_threshold=upper_threshold)                                        
        X_new = scaler.fit_transform(X)        
        X_new_norm = np.linalg.norm(X_new)
        assert X_new_norm>=lower_threshold and \
               X_new_norm<=upper_threshold, \
                   "Scaling didn't work. X_new_norm = {n}".format(
                   n=str(X_new_norm))        
        X_old = scaler.inverse_transform(X_new)
        X_old_norm = np.linalg.norm(X_old)

        assert np.isclose(X_orig_norm, X_old_norm), \
            "Reverse transform didn't work\
                \nX_orig_norm = {n1}\nX_old_norm={n2}".format(n1=str(X_orig_norm),
                n2=str(X_old_norm))
        
@mark.utils
@mark.data_manager
@mark.gradient_scaler
@mark.gradient_scaler_2d
def test_gradient_scaler_2d():            
    lower_threshold = 1e-10
    upper_threshold = 1e10
    lows = [1e-20, 1e15, 1] 
    highs = [1e-10, 1e20, 5]
    for g in zip(lows, highs):    
        X = np.random.default_rng().uniform(low=g[0], high=g[1], size=(20,4))                
        X_orig_norm = (np.linalg.norm(X))                
        scaler = GradientScaler(lower_threshold=lower_threshold, 
                                upper_threshold=upper_threshold)                                        
        X_new = scaler.fit_transform(X)        
        X_new_norm = np.linalg.norm(X_new)
        assert X_new_norm>=lower_threshold and \
               X_new_norm<=upper_threshold, \
                   "Scaling didn't work. X_new_norm = {n}".format(
                   n=str(X_new_norm))        
        X_old = scaler.inverse_transform(X_new)
        X_old_norm = np.linalg.norm(X_old)

        assert np.allclose(X_orig_norm, X_old_norm), \
            "Reverse transform didn't work\
                \nX_orig_norm = {n1}\nX_old_norm={n2}".format(n1=str(X_orig_norm),
                n2=str(X_old_norm))

# --------------------------------------------------------------------------  #
#                       TEST MINMAX SCALER                                    #
# --------------------------------------------------------------------------  #
@mark.utils
@mark.data_manager
@mark.minmax
def test_minmax_scaler():
    x = np.array([[0,0,22],
                [0,1,17],
                [0,1,2]], dtype=float)
    x_new = np.array([[0,0,1],
                    [0,1,15/20],
                    [0,1,0]], dtype=float)
    scaler = MinMaxScaler()
    x_t = scaler.fit_transform(x)
    assert np.array_equal(x_new, x_t), "Minmax scaler not working"    
# --------------------------------------------------------------------------  #
#                        TEST DATA SPLIT                                      #
# --------------------------------------------------------------------------  #  
@mark.utils
@mark.data_manager
@mark.data_split  
def test_data_split():
    X, y = make_classification(n_classes=4, n_informative=3)
    X_train, X_test, y_train, y_test = data_split(X,y, stratify=True)
    n_train = y_train.shape[0]
    n_test = y_test.shape[0]
    train_values, train_counts = np.unique(y_train, return_counts=True)
    test_values, test_counts = np.unique(y_test, return_counts=True)
    train_proportions = train_counts / n_train
    test_proportions = test_counts / n_test
    assert np.allclose(train_proportions, test_proportions, rtol=0.1), "Data split stratification problem "

# --------------------------------------------------------------------------  #
#                        TEST UNPACK PARAMETERS                               #
# --------------------------------------------------------------------------  #  
@mark.utils
@mark.unpack_parameters
def test_unpack_parameters():
    X = np.random.rand(5)
    bias, weights = unpack_parameters(X)
    assert bias == X[0], "Bias not unpacked correctly"
    assert np.array_equal(weights, X[1:]), "weights not unpacked correctly"
    X = np.random.rand(5,3)
    bias, weights = unpack_parameters(X)
    assert bias.shape == (3,), "Bias not unpacked correctly"
    assert np.array_equal(bias, X[0,:]), "Bias not unpacked correctly "
    assert weights.shape == (4,3), "Weights not unpacked correctly"
    assert np.array_equal(weights, X[1:,:]), "Weights not unpacked correctly"






# %%



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.14                                                            #
# File    : text.py                                                           #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Sunday, June 14th 2020, 11:34:29 pm                         #
# Last Modified : Sunday, June 14th 2020, 11:56:46 pm                         #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Text-based visualizations."""
from abc import ABC, abstractmethod
from collections import OrderedDict
import itertools 
import numpy as np
import pandas as pd
from tabulate import tabulate

from mlstudio.utils.format import proper
from mlstudio.supervised.observers.early_stop import Performance
from mlstudio.utils.print import Printer

class Summary(ABC):
    """Base class for all optimization summary classes."""

    def __init__(self, model):
        self.model = model
        self._printer = Printer()

    @abstractmethod
    def report(self):
        pass

class OptimizationSummary(Summary):
    """Reports summary information for an optimization."""

    def _extract_data(self):
        """Extracts required data from the model."""
        bb = self.model.blackbox_
        data = {}
        data['start'] = bb.start
        data['end'] = bb.end
        data['duration'] = bb.duration
        data['epochs'] = bb.total_epochs
        data['batches'] = bb.total_batches
        return data

    def report(self):
        data = self._extract_data()
        optimization_summary = {'Name': self.model.description,
                                'Start': str(data['start']),
                                'End': str(data['end']),
                                'Duration': str(data['duration']) + " seconds.",
                                'Epochs': str(data['epochs']),
                                'Batches': str(data['batches'])}
        self._printer.print_dictionary(optimization_summary, "Optimization Summary")          

class OptimizationPerformance(Summary):
    """Reports performance information for an optimization."""

    def _extract_data(self):
        """Extracts required data from the model."""
        final_results = self.model.final_results_
        try:
            best_results = self.model.best_results_ 
        except:
            best_results = {}
        return final_results, best_results 

    def _report(self, result, best_or_final):
        datasets = {'train': 'Training', 'val': 'Validation'}
        keys = ['train', 'val']
        metrics = ['cost', 'score']
        print_data = []
        # If printing 'best_results', extract the epoch
        if best_or_final == 'best':
            d = {'label': 'Best Epoch', 'data' : str(result.get('epoch'))}
            print_data.append(d)
        # Format labels and data for printing from result parameter
        for performance in list(itertools.product(keys, metrics)):
            d = {}
            key = performance[0] + '_' + performance[1]
            if result.get(key):
                label = proper(best_or_final) + ' ' + datasets[performance[0]] \
                    + ' ' + proper(performance[1]) 
                d['label'] = label
                if performance[1] == 'score' and hasattr(self.model, 'scorer'):                    
                    d['data'] = str(np.round(result.get(key),4)) + " " + self.model.scorer.name
                else:
                    d['data'] = str(np.round(result.get(key),4)) 
                print_data.append(d) 

        # Print performance statistics
        performance_summary = OrderedDict()
        for i in range(len(print_data)):
            performance_summary[print_data[i]['label']] = print_data[i]['data']
        title = proper(best_or_final) + " Weights Performance Summary"
        self._printer.print_dictionary(performance_summary, title)        

    def report(self):
        final_results, best_results  = self._extract_data()
        self._report(final_results, 'final')
        if best_results:
            self._report(best_results, 'best')

class OptimizationCriticalPoints(Summary):
    """Prints statistics at critical points during the optimization.
    
    Critical points are when the optimization changed state from stabilized
    to not stabilized.
    """
    def _extract_data(self):
        critical_points = None
        if hasattr(self.model, 'critical_points_'):
            critical_points = self.model.critical_points_
        return critical_points

    def report(self):
        critical_points = self._extract_data()
        if critical_points:        
            cp = []
            for p in critical_points:
                d = {}
                for k,v in p.items():
                    d[proper(k)] = v
                cp.append(d)                      
            self._printer.print_title("Critical Points")
            df = pd.DataFrame(cp) 
            df = df.drop(['Theta', 'Gradient'], axis=1)
            df.set_index('Epoch', inplace=True)
            print(tabulate(df, headers="keys"))
            print("\n")       


class OptimizationHyperparameters(Summary):
    """Reports the hyperparameters used for the optimization."""

    def report(self):
        """Renders report of hyperparameters used for the optimization."""
        hyperparameters = OrderedDict()
        def get_params(o):
            params = o.get_params()
            for k, v in params.items():
                if isinstance(v, (str, dict, bool, int, float, np.ndarray, \
                    np.generic, list)) or v is None:
                    k = o.__class__.__name__ + '__' + k
                    hyperparameters[k] = str(v)
                else:
                    get_params(v)
        get_params(self.model)

        self._printer.print_dictionary(hyperparameters, "Model HyperParameters")             


class OptimizationFeatures(Summary):
    """Reports the parameters with feature names if the feature names are available."""

    def __init__(self, model):
        super(OptimizationFeatures, self).__init__(
            model = model
        )            

    def report(self):
        theta = OrderedDict()
        theta['Intercept'] = str(np.round(self.model.intercept_, 4))      

        # If no features were provided to the estimator, create dummy features.
        if self.model.feature_names is None:
            self.features = []
            for i in np.arange(len(self.model.coef_)):
                self.features.append("Feature_" + str(i))

        for k, v in zip(self.features, self.model.coef_):
            theta[k]=str(np.round(v,4))  
        self._printer.print_dictionary(theta, "Model Parameters")        

class OptimizationReport:
    """Prints and optimization report. 

    Parameters
    ----------
    reports : list default=['summary', 'performance', 'critical_points',
                            'features', 'hyperparameters']
        The reports in the order to be rendered. The valid report names are:
            'summary' : prints summary data for optimzation
            'hyperparameters' : prints the hyperparameters used for the optimization
            'performance' : prints performance in terms of cost and scores
            'critical_points' : prints cost and scores at critical points 
                during the optimization
            'features' : prints the best or final intercept and coeficients 
                by feature name if feature names are available. Best results 
                are printed if the Performance observer is used and the 
                'best_or_final' parameter = 'best'. Otherwise, final results
                will be printed.

    """
    def __init__(self, model, reports=None):
        self.model = model
        self.reports = reports

    def report(self):

        valid_reports = ['summary', 'performance', 'critical_points', 
                           'features', 'hyperparameters']

        default_reports = ['summary', 'performance', 
                           'features', 'hyperparameters']                           

        reports = {'summary': OptimizationSummary(model=self.model), 
                   'performance': OptimizationPerformance(model=self.model),
                   'critical_points': OptimizationCriticalPoints(model=self.model), 
                   'features' : OptimizationFeatures(model=self.model), 
                   'hyperparameters' : OptimizationHyperparameters(model=self.model)}

        if self.reports is None:
            self.reports = default_reports
        if set(self.reports).issubset(set(default_reports)):            
            for report in self.reports:
                reports[report].report()
        else:
            msg = "One of your reports is not a valid report. Valid reports\
                 include {r}".format(str(default_reports))
            raise ValueError(msg)


def get_performance_observer(model):
    """Obtains performance observer if it was used."""
    observers = model.observers
    if observers:
        for name, observer in observers.items():
            if isinstance(observer, Performance):
                return observer
    return None        




#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# ============================================================================ #
# Project : MLStudio                                                           #
# Version : 0.1.0                                                              #
# File    : data_manager.py                                                    #
# Python  : 3.8.2                                                              #
# ---------------------------------------------------------------------------- #
# Author  : John James                                                         #
# Company : DecisionScients                                                    #
# Email   : jjames@decisionscients.com                                         #
# URL     : https://github.com/decisionscients/MLStudio                        #
# ---------------------------------------------------------------------------- #
# Created       : Sunday, March 15th 2020, 6:52:47 pm                          #
# Last Modified : Sunday, March 15th 2020, 6:57:11 pm                          #
# Modified By   : John James (jjames@decisionscients.com)                      #
# ---------------------------------------------------------------------------- #
# License : BSD                                                                #
# Copyright (c) 2020 DecisionScients                                           #
# ============================================================================ #
#%%
"""Data manipulation functions."""
from abc import ABC, abstractmethod
from itertools import combinations_with_replacement
from math import ceil, floor

import numpy as np
from numpy.random import RandomState
import pandas as pd
from scipy.sparse import isspmatrix_coo, issparse, csr_matrix, hstack
from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.utils import shuffle, check_array
from sklearn.preprocessing import LabelBinarizer, LabelEncoder

from mlstudio.utils.validation import check_X_y
# --------------------------------------------------------------------------- #
#                           DATA PREPARATION                                  #
# --------------------------------------------------------------------------- #

def check_coo(X):
    """Converts coo matrices to csr format."""
    if issparse(X):
        if isspmatrix_coo(X):
            X = X.tocsr()
    return X    
# --------------------------------------------------------------------------- #
def unpack_parameters(theta):
    """Unpacks the parameters theta and returns bias and weights."""
    if np.ndim(theta) == 0:
        bias = np.array([0])
        weights = np.array(theta)
    elif np.ndim(theta) == 1:
        bias = theta[0]
        weights = theta[1:]
    else:
        bias = theta[0,:]
        weights = theta[1:,:]
    return bias, weights
# --------------------------------------------------------------------------- #
def encode_labels(y):
    """Encodes labels to have values from 0 to n_classes - 1
    
    Parameters
    ----------
    y : array shape (n_samples,)
        Labels to encode

    Returns
    -------
    y : Encoded labels with values from 0 to n_classes - 1
    """
    classes = np.sort(np.unique(y))
    n_classes = len(classes)
    binary_classes = np.arange(n_classes)
    if np.array_equal(classes, binary_classes):
        return y
    else:
        encoder = LabelEncoder()
        return encoder.fit_transform(y)

# --------------------------------------------------------------------------- #
#                               TRANSFORMERS                                  #
# --------------------------------------------------------------------------- #
class AddBiasTerm(BaseEstimator, TransformerMixin):
    """Adds bias term of ones to matrix."""

    def fit(self, X, y=None):
        """Fits data to the transformer."""
        return self

    def _transform_numpy(self, X):
        """Adds bias term to numpy matrix."""
        return np.insert(X, 0, 1.0, axis=1)
    
    def _transform_csr(self, X):
        """Adds bias term to csr matrix."""
        X = check_coo(X)
        ones = np.ones((X.shape[0],1))
        bias_term = csr_matrix(ones, dtype=float)
        X = hstack((bias_term, X)) 
        X = check_coo(X)
        return X

    def transform(self, X, y=None):
        """Adds bias term to matrix and returns it to the caller."""
        X = check_array(X, accept_sparse=True, accept_large_sparse=True)
        if issparse(X):
            X = self._transform_csr(X)
        else:
            X = self._transform_numpy(X)
        return X        
    
    def inverse_transform(self, X):
        """Removes bias term from matrix and returns it to caller."""
        X = check_coo(X)
        return X[:,1:]
                        
# --------------------------------------------------------------------------- #
class ZeroBiasTerm(BaseEstimator, TransformerMixin):
    """Zeros out bias term in a parameters matrix or tensor."""

    def fit(self, X, y=None):
        """Fits data to the transformer."""
        return self

    def transform(self, X):
        """Zero out bias term in numpy matrix."""
        if np.ndim(X) == 1:
            X[0] = 0
        else:
            X[0,:] = np.zeros(shape=X.shape[1])
        return X
           

# --------------------------------------------------------------------------- #
class DataProcessor(ABC, BaseEstimator):
    """Prepares data for training.
    
    Parameters
    ----------
    val_size : float (default = 0.3)
        The proportion of data to be allocated to a validation set.

    random_state : int or None (default = None)
        The seed for pseudo randomization.
    
    """
    def __init__(self, estimator, random_state=None):        
        self.estimator = estimator
        self.random_state = random_state


    def fit(self, X, y=None):
        """Returns self."""
        return self

    @abstractmethod
    def transform(self, X, y=None):
        """Processes data and returns a dictionary to the calling object.
        
        Parameters
        ----------
        X : array_like of shape (n_samples,n_features)
            The matrix of inputs 

        y : array_like of shape(n_samples,)
            An array of target values.

        Returns
        -------
        dict : Dictionary containing transformed data.           

        """
        pass

    def fit_transform(self, X, y=None):
        """Calls fit and transform."""
        return self.fit(X, y).transform(X, y)        

# --------------------------------------------------------------------------- #
class RegressionDataProcessor(DataProcessor):
    """Prepares data for regression."""

    def transform(self, X, y):
        """Prepares data for regression."""
        X_train, y_train = check_X_y(X, y)
        X_train = AddBiasTerm().fit_transform(X_train)
        # Initialize data dictionary and n_features (including bias term)
        d = {}
        d['n_features_'] = X_train.shape[1]        
        # Return training and validation data if val_size is truthy
        if self.estimator.val_size:
            X_train, X_val, y_train, y_val = data_split(X=X_train, y=y_train, 
                        stratify=False, test_size=self.estimator.val_size, 
                        random_state=self.random_state)
            d['X_train_'] = X_train
            d['y_train_'] = y_train
            d['X_val_'] = X_val
            d['y_val_'] = y_val            
        # Otherwise return just training data
        else:             
            d['n_features_'] = X_train.shape[1]
            d['X_train_'] = X_train
            d['y_train_'] = y_train

        return d

# --------------------------------------------------------------------------- #
class LogisticRegressionDataProcessor(DataProcessor):
    """Prepares data for regression."""

    def transform(self, X, y):
        """Prepares data for regression."""
        # Check the data and add bias term
        X_train, y_train = check_X_y(X, y)
        X_train = AddBiasTerm().fit_transform(X_train)
        # Initialize data dictionary and n_features (including bias term)
        d = {}
        d['n_features_'] = X_train.shape[1]        
        # Return stratified training and validation data if val_size is truthy
        if self.estimator.val_size:
            X_train, X_val, y_train, y_val = data_split(X=X_train, y=y_train, 
                        stratify=True, test_size=self.estimator.val_size, 
                        random_state=self.random_state)
            d['X_train_'] = X_train
            d['y_train_'] = y_train
            d['X_val_'] = X_val
            d['y_val_'] = y_val            
        # Otherwise return just training data
        else:             
            d['n_features_'] = X_train.shape[1]
            d['X_train_'] = X_train
            d['y_train_'] = y_train

        return d
# --------------------------------------------------------------------------- #
class MulticlassDataProcessor(DataProcessor):
    """Prepares data for multi-class classification."""

    def transform(self, X, y):
        """Prepares data for multi-class classification."""
        if len(np.unique(y)) < 3:
            msg = "This data processor is for multi-class settings with classes\
                >2. Please use the LogisticRegressionDataProcessor instead."
            raise Exception(msg)

        # Check the data, add bias term, and encode y to values [0,n_classes - 1]
        X_train, y_train = check_X_y(X, y)
        X_train = AddBiasTerm().fit_transform(X_train)
        y_train = encode_labels(y_train)
        # Initialize data dict, n_features (including bias term) and n_classes
        d = {}
        d['n_features_'] = X_train.shape[1]  
        d['n_classes_'] = len(np.unique(y_train))
        # Return stratified training and validation data if val_size is truthy
        if self.estimator.val_size:
            X_train, X_val, y_train, y_val = data_split(X=X_train, y=y_train, 
                        stratify=True, test_size=self.estimator.val_size, 
                        random_state=self.random_state)
            d['X_train_'] = X_train
            d['y_train_'] = y_train            
            d['X_val_'] = X_val
            d['y_val_'] = y_val
        # Otherwise return training data
        else:
            d['X_train_'] = X_train
            d['y_train_'] = y_train        
        return d                               

# --------------------------------------------------------------------------- #
class NormScaler(TransformerMixin, BaseEstimator):
    """Scalers a vector to unit length.  

    Scaling a sample 'x' to 0-1 is calculated as:

        X_new = X/ X.norm

    Note: Works for dense matrices only.

    Attributes
    ----------
    r : float
        The magnitude of the vector

    """        

    def __init__(self, clip_norm=1):        
        self.clip_norm = clip_norm
        self.r_ = None

    def fit(self, X, y=None):
        """Computes the Frobenius norm of the input vector
        
        Parameters
        ----------
        X : array-like, shape [n_features,]
            The data used to compute the mean and standard deviation
            used for centering and scaling.

        y : Ignored

        """
        self.r_= np.linalg.norm(X)
        return self

    def transform(self, X):
        """Scales features to have a norm of 1

        Parameters
        ----------
        X : array-like, shape [n_features,]
            The data to scale

        Returns
        -------
        Xt : array-like of same shape as X
        """
        X = np.divide(X, self.r_) * self.clip_norm               
        return X

    def fit_transform(self, X):
        """Combines fit and transform methods.
        
        Parameters
        ----------
        X : array-like, shape [n_features,]
            The data to scale
        
        Returns
        -------
        Xt : array-like of same shape as X
        """
        self.fit(X)
        return self.transform(X)

    def inverse_transform(self, X):
        """Inverses the standardization process.

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The centered and scaled data.

        Returns
        -------
        array-like of same shape as X, with data returned to original
        un-standardized values.
        """
        X = X * self.r_ / self.clip_norm    
        return X
# --------------------------------------------------------------------------- #

class MinMaxScaler(TransformerMixin, BaseEstimator):
    """Scales each feature to values between 0 and 1.

    Scaling a sample 'x' to 0-1 is calculated as:

        X_new = (X - X_min) / (X_max-X_min)

    Note: Works for dense matrices only.

    Attributes
    ----------
    data_min_ : ndarray, shape (n_features)
        Per feature minimum seen in the data

    data_max_ : ndarray, shape (n_features)
        Per feature maximum seen in the data        

    data_range_ : ndarray, shape (n_features)
        Per feature range ``(data_max_ - data_min_)`` seen in the data

    """        

    def __init__(self):        
        pass

    def fit(self, X, y=None):
        """Computes the min and max on X for scaling.
        
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data used to compute the mean and standard deviation
            used for centering and scaling.

        y : Ignored

        """
        self.data_min_ = np.amin(X, axis=0)
        self.data_max_ = np.amax(X, axis=0)
        self.data_range_ = self.data_max_ - self.data_min_
        return self

    def transform(self, X):
        """Scales features to range 0 to 1.

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data to center and scale.

        Returns
        -------
        Xt : array-like of same shape as X
        """
        X = X - self.data_min_
        X = np.divide(X, self.data_range_, 
                      out = np.zeros(X.shape,dtype=float), 
                      where = self.data_range_ != 0)        
        return X

    def fit_transform(self, X):
        """Combines fit and transform methods.
        
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data to center and scale.
        
        Returns
        -------
        Xt : array-like of same shape as X
        """
        self.fit(X)
        return self.transform(X)

    def inverse_transform(self, X):
        """Inverses the standardization process.

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The centered and scaled data.

        Returns
        -------
        array-like of same shape as X, with data returned to original
        un-standardized values.
        """
        X = X * self.data_range_
        X = X + self.data_min_
        return X
# --------------------------------------------------------------------------- #
class StandardScaler(TransformerMixin, BaseEstimator):
    """Standardizes data to a zero mean and unit variance.

    Standardizing a sample 'x' is calculated as:

        z = (x - u) / s
    where 'u' is either the mean of the training samples 'x', or zero of 
    'center=False' and 's' is the standard deviation of the training samples
    or one if 'scale=False'.

    Parameters
    ----------
    center : Bool, optional (default=True)
        If True, center the data by subtracting the means of the variables.

    scale : Bool, optional (default=True)
        If True, scale the data to a unit variance.

    Attributes
    ----------
    mean_ : array-like, shape (n_features)
        The mean value for each feature in the training set.
        Equal to zero if 'center=False'.

    std_ : array-like, shape (n_features)
        The standard deviation for each feature in the training set.
        Equal to one if 'scale=False'.
    """        

    def __init__(self, center=True, scale=True):
        self.center = center
        self.scale = scale
        self.mean_=0
        self.std_=1

    def fit(self, X, y=None):
        """Computes the mean and std for centering and scaling the data.
        
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data used to compute the mean and standard deviation
            used for centering and scaling.

        y : Ignored

        """
        if self.center:
            self.mean_ = np.mean(X,axis=0)
        if self.scale:
            self.std_ = np.std(X,axis=0)
        return self

    def transform(self, X):
        """Center and scale the data.

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data to center and scale.

        Returns
        -------
        array-like of same shape as X, centered and scaled
        """
        z = (X-self.mean_)/self.std_
        return z

    def inverse_transform(self, X):
        """Inverses the standardization process.

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The centered and scaled data.

        Returns
        -------
        array-like of same shape as X, with data returned to original
        un-standardized values.
        """
        X = X * self.std_
        X = X + self.mean_
        return X

    def fit_transform(self, X):
        """Calls fit and transform methods."""
        self.fit(X)
        return self.transform(X)

# --------------------------------------------------------------------------  #
#                           GRADIENT SCALING                                  #        
# --------------------------------------------------------------------------  #        
class GradientScaler(BaseEstimator, TransformerMixin):
    """Scales and/or normalizes exploding and vanishing gradients. 

    If the norm of the gradient is below the lower threshold, the gradient
    is rescaled to a unit vector. If it is greater than the upper threshold
    then the new gradient is given by:

    ..math:: \nabla_{new} = \frac{\text{upper_threshold}}{\lVert \nabla \rVert} \times \nabla
    
    Parameters
    ----------
    lower_threshold : float (default=1e-15)
        The lower threshold for the magnitude of the vector.

    upper_threshold : float (default=1e15)
        The upper threshold for the magnitude of the vector.   

    """

    def __init__(self, lower_threshold=1e-10, upper_threshold=1e10): 
        self.lower_threshold  = lower_threshold
        self.upper_threshold = upper_threshold
        self.normalizer_ = None

    def fit(self, X, y=None):
        """Fits the transformer to the data. """  
        self._r = np.linalg.norm(X)
        return self       

    def transform(self, X):
        """Transforms the data."""                
        if self._r < self.lower_threshold:
            X = X  * self.lower_threshold / self._r                        
        elif self._r > self.upper_threshold:
            X = X  * self.upper_threshold / self._r            
        return X
            
    def fit_transform(self, X):
        """Performs fit and transform."""
        self.fit(X)
        return self.transform(X)

    def inverse_transform(self, X):
        """Apply the inverse transformation."""
        if self._r < self.lower_threshold:
            X = X  * self._r / self.lower_threshold                         
        elif self._r > self.upper_threshold:
            X = X  * self._r / self.upper_threshold                    
        return X

# --------------------------------------------------------------------------- #
#                            SHUFFLE DATA                                     #
# --------------------------------------------------------------------------- #
def shuffle_data(X, y=None, random_state=None):
    """ Random shuffle of the samples in X and y.
    
    Shuffles data

    Parameters
    ----------
    X : array_like of shape (m, n_features)
        Input data

    y : array_like of shape (m,)
        Target data    

    Returns
    -------
    Shuffled X, and y
    
    """    
    X, y = shuffle(X, y, random_state=random_state)
    return X, y

# --------------------------------------------------------------------------- #
#                              SAMPLE                                         #
# --------------------------------------------------------------------------- #    
def sampler(X, y, size=1, replace=True, random_state=None):
    """Generates a random sample of a given size from a data set.

        Parameters
    ----------
    X : array_like of shape (m, n_features)
        Input data

    y : array_like of shape (m,)
        Target data    

    size : int
        The size of the dataset.

    replace : Bool. 
        Whether to sample with or without replacement

    random_state : int
        random_state for reproducibility
    
    Returns
    -------
    X, y    : Random samples from data sets X, and y of the designated size.

    """
    import numpy as np

    nobs = X.shape[0]
    idx = np.random.choice(a=nobs, size=size, replace=True)
    return X[idx], y[idx]

# --------------------------------------------------------------------------- #
#                            SPLIT DATA                                       #
# --------------------------------------------------------------------------- #
def data_split(X, y, test_size=0.3, shuffle=False, stratify=False, random_state=None):
    """ Split the data into train and test sets 
    
    Splits inputs X, and y into training and test sets of proportions
    1-test_size, and test_size respectively.

    Parameters
    ----------
    X : array_like of shape (m, n_features)
        Input data

    y : array_like of shape (m,)
        Target data

    test_size : float, optional (default=0.3)
        The proportion of X, and y to be designated to the test set.

    shuffle : bool, optional (default=True)
        Bool indicating whether the data should be shuffled prior to split.

    stratify : bool, optional (default=False)
        If True, stratified sampling is performed. 

    random_state : int, optional (default=None)
        Random state variable

    Returns
    -------
    X_train : array-like
        Training data 

    X_test : array-like
        Test data

    y_train : array-like
        Targets for X_train

    y_test : array_like
        Targets for X_test 
    """
    if isspmatrix_coo(X):
        X = X.tocsr()
    if isspmatrix_coo(y):
        y = y.tocsr()
        
    if X.shape[0] != y.shape[0]:
        raise ValueError("X and y have incompatible shapes. Expected "
                         "X.shape[0]=y.shape[0] however X.shape[0] = %d "
                         " and y.shape[0] = %d." % (X.shape[0], y.shape[0]))

    if not stratify:
        if shuffle:
            X, y = shuffle_data(X, y, random_state)
        split_i = len(y) - int(len(y) // (1 / test_size))
        X_train, X_test = X[:split_i], X[split_i:]
        y_train, y_test = y[:split_i], y[split_i:]
    else:
        train_idx = []
        test_idx = []
        classes = np.unique(y)
        for k in classes:
            # Obtain the indices and number of samples for class k
            idx_k = np.array(np.where(y == k)).reshape(-1,1)  
            n_samples_k = idx_k.shape[0]
            # Compute number of training and test samples
            n_train_samples_k = ceil(n_samples_k * (1-test_size))
            n_test_samples_k = n_samples_k - n_train_samples_k
            # Shuffle the data
            if shuffle:
                if random_state:
                    np.random.random_state(random_state)
                np.random.shuffle(idx_k)
            # Allocate corresponding indices to training and test set indices
            train_idx_k = idx_k[0:n_train_samples_k]
            test_idx_k = idx_k[n_train_samples_k:n_train_samples_k+n_test_samples_k]
            # Maintain indices in a list
            train_idx.append(train_idx_k)
            test_idx.append(test_idx_k)
        # Concatenate all indices into a training and test indices
        train_idx = np.concatenate(train_idx).ravel()
        test_idx = np.concatenate(test_idx).ravel()
        # Slice and dice.
        y_train, y_test = y[train_idx], y[test_idx]
        X_train, X_test = X[train_idx], X[test_idx]

    return X_train, X_test, y_train, y_test

def batch_iterator(X, y=None, batch_size=None):
    """Batch generator.
    
    Creates an iterable of batches of the designated batch size.

    Parameters
    ----------
    X : array-like
        A features matrix of shape (m, n_features), where m is the number of 
        examples in X.
    
    y : array-like, optional (default=None)
        The target vector of shape (m,)

    batch_size : None or int, optional (default=None)
        The number of observations to be included in each batch. 

    Returns
    -------
    array-like
        Returns inputs in batches of batch_size. If batch_size
        is None, a single batch containing all data is generated.
    
    """
    n_samples = X.shape[0]
    if batch_size is None:
        batch_size = n_samples    
    for i in np.arange(0, n_samples, batch_size):
        if y is not None:
            yield X[i:i+batch_size], y[i:i+batch_size]
        else:
            yield X[i:i+batch_size]

def one_hot(x, n_classes=None, dtype='float32'):
    """Converts a vector of integers to one-hot encoding. 
    
    Creates a one-hot matrix for multi-class classification
    and categorical cross_entropy.

    Parameters
    ----------
    x : array-like of shape(n_observations,)
        Vector of integers (from 0 to num_classes-1) to be converted to one-hot matrix.
    n_classes : int
        Number of classes
    dtype : Data type to be expected, as a string
        ('float32', 'float64', 'int32', ...)

    Returns
    -------
    A binary one-hot matrix representation of the input. The classes axis
    is placed last.
    """
    x = np.array(x, dtype='int')
    if not n_classes:
        n_classes = np.amax(x) + 1
    one_hot = np.zeros((x.shape[0], n_classes))
    one_hot[np.arange(x.shape[0]), x] = 1
    return one_hot

def todf(x, stub):
    """Converts nested array to dataframe."""
    n = len(x[0])
    df = pd.DataFrame()
    for i in range(n):
        colname = stub + str(i)
        vec = [item[i] for item in x]
        df_vec = pd.DataFrame(vec, columns=[colname])
        df = pd.concat([df, df_vec], axis=1)
    return(df)  



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : learning_rate.py                                                  #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Friday, May 15th 2020, 9:48:31 pm                           #
# Last Modified : Friday, May 15th 2020, 9:48:31 pm                           #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Learning rate schedules."""
from abc import abstractmethod
import math
import numpy as np
from tabulate import tabulate

from mlstudio.supervised.observers.base import Observer, PerformanceObserver
from mlstudio.supervised.core.scorers import MSE
from mlstudio.utils.validation import validate_bool
from mlstudio.utils.validation import validate_int
from mlstudio.utils.validation import validate_objective, validate_optimizer
from mlstudio.utils.validation import validate_scorer, validate_string
from mlstudio.utils.validation import validate_zero_to_one, validate_int
from mlstudio.utils.validation import validate_metric
# --------------------------------------------------------------------------  #
class LearningRateSchedule(Observer):
    """Base class for learning rate schedules. 
    
    Parameters
    ----------
    eta0 : float (default=0.1)
        The initial learning rate that is decayed during optimization. This 
        will override the eta0 parameter on the estimator.

    eta_min : float(default=1e-4)
        The minimum allowable learning rate

    decay_factor : float (default=0.5) 
        The factor by which the learning rate is decayed

    """

    @abstractmethod
    def __init__(self, eta0=0.1, 
                       eta_min=1e-4,
                       decay_factor=0.5):    
        super(LearningRateSchedule, self).__init__()
        self.eta0 = eta0
        self.eta_min = eta_min  
        self.decay_factor = decay_factor        

    def _validate(self):
        """Performs validation of hyperparameters."""
        validate_zero_to_one(param=self.eta0, 
                             param_name='eta0', 
                             left='open', right='open')
        validate_zero_to_one(param=self.eta_min, 
                             param_name='eta_min', 
                             left='open', right='open')
        validate_zero_to_one(param=self.decay_factor, 
                             param_name='decay_factor',
                             left='closed', right='open')

    @abstractmethod
    def _compute_learning_rate(self, epoch, log):
        pass

    def on_train_begin(self, log=None):
        super(LearningRateSchedule, self).on_train_begin(log)        
        self._validate()
        self.model.eta = self.eta0

    def on_epoch_end(self, epoch, log=None):
        super(LearningRateSchedule, self).on_epoch_begin(epoch=epoch, log=log)   
        self.model.eta = max(self.eta_min,\
            self._compute_learning_rate(epoch=epoch, log=log))
    
        

# --------------------------------------------------------------------------  #
class StepDecay(LearningRateSchedule):
    """ Time decay learning rate schedule as:
    .. math:: \eta_0 \times \gamma^{\text{floor((1+epoch)/decay_steps)}}
    
    Parameters
    ----------
    eta0 : float (default=0.1)
        The initial learning rate that is decayed during optimization. This 
        will override the eta0 parameter on the estimator.

    eta_min : float(default=1e-4)
        The minimum allowable learning rate        
         
    decay_factor : float (default= 0.5)
        The factor used as the base of the polynomial used to compute
        the decayed learning rate.

    decay_steps : int (default=100)
        The total number of steps to decay the rate from the initial
        learning rate to the minimum learning rate.

    """

    def __init__(self, eta0=0.1, 
                       eta_min=1e-4,
                       decay_factor=0.5,
                       decay_steps=10):        
        super(StepDecay, self).__init__(
            eta0=eta0,
            eta_min=eta_min,
            decay_factor=decay_factor)              

        self.name = "Step Decay Learning Rate Schedule"
        self.decay_steps = decay_steps

    def _validate(self):
        """Performs hyperparameter validation """
        super(StepDecay, self)._validate()
        validate_int(param=self.decay_steps, 
                     param_name='decay_steps',
                     minimum=1, left='closed', right='open')                

    def _compute_learning_rate(self, epoch, log):        
        exponent = (1 + epoch) // self._steps
        return self.eta0 * \
            np.power(self.decay_factor, exponent)

    def on_train_begin(self, log=None):
        super(StepDecay, self).on_train_begin(log=log)        
        self.decay_steps = min(self.decay_steps, self.model.epochs)            
        self._steps = self.model.epochs // self.decay_steps

# --------------------------------------------------------------------------  #
class TimeDecay(LearningRateSchedule):
    """ Time decay learning rate schedule as:
    .. math:: \eta_t=\frac{\eta_0}{1+b\cdot t} 

    Parameters
    ----------
    eta0 : float (default=0.1)
        The initial learning rate that is decayed during optimization. This 
        will override the eta0 parameter on the estimator.

    eta_min : float(default=1e-4)
        The minimum allowable learning rate        

    decay_factor : float or 'optimal' (default= 'optimal')
        If 'optimal', the decay rate will be computed based upon the 
        learning rate and the anticipated number of epochs

    """

    def __init__(self, eta0=0.1, 
                       eta_min=1e-4,
                       decay_factor='optimal'):        
        super(TimeDecay, self).__init__(
            eta0=eta0,
            eta_min=eta_min,
            decay_factor=decay_factor)

        self.name = "Time Decay Learning Rate Schedule"              

    def _compute_optimal_decay_factor(self):
        return 1 - (self.eta0 - self.eta_min)\
            / self.model.epochs

    def _compute_learning_rate(self, epoch, log):
        return self.eta0 / \
            (1 + self.decay_factor * epoch)

    def on_train_begin(self, log=None):
        if self.decay_factor == 'optimal':
            self.decay_factor = self._compute_optimal_decay_factor()            
        super(TimeDecay, self).on_train_begin(log)        

# --------------------------------------------------------------------------  #
class SqrtTimeDecay(LearningRateSchedule):
    """ Time decay learning rate schedule as:
    .. math:: \eta_t=\frac{\eta_0}{1+b\cdot \sqrt{t}} 

    Parameters
    ----------
    eta0 : float (default=0.1)
        The initial learning rate that is decayed during optimization. This 
        will override the eta0 parameter on the estimator.

    eta_min : float(default=1e-4)
        The minimum allowable learning rate        

    decay_factor : float (default= 0.5)
        The factor used as to decay the learning rate.
    """
    def __init__(self, eta0=0.1, 
                       eta_min=1e-4,
                       decay_factor=0.5):        
        super(SqrtTimeDecay, self).__init__(
            eta0=eta0,
            eta_min=eta_min,
            decay_factor=decay_factor)              

        self.name = "Sqrt Time Decay Learning Rate Schedule"

    def _compute_learning_rate(self, epoch, log):
        return self.eta0 / \
            (1 + self.decay_factor * np.sqrt(epoch))       
               


# --------------------------------------------------------------------------  #
class ExponentialDecay(LearningRateSchedule):
    """ Exponential decay learning rate schedule as:
    .. math:: \eta_t=\eta_0 \cdot \text{exp}(-b\cdot t)

    Parameters
    ----------
    eta0 : float (default=0.1)
        The initial learning rate that is decayed during optimization. This 
        will override the eta0 parameter on the estimator.

    eta_min : float(default=1e-4)
        The minimum allowable learning rate        

    decay_factor : float (default= 0.1)
        The decay factor used in the exponent for the learning rate computation
    """

    def __init__(self, eta0=0.1, 
                       eta_min=1e-4,
                       decay_factor=0.1):        
        super(ExponentialDecay, self).__init__(
            eta0=eta0,
            eta_min=eta_min,
            decay_factor=decay_factor)

        self.name = "Exponential Decay Learning Rate Schedule"

    def _compute_learning_rate(self, epoch, log):
        return self.eta0 * \
            np.exp(-self.decay_factor * epoch)

# --------------------------------------------------------------------------  #
class ExponentialStepDecay(LearningRateSchedule):
    """ Exponential step decay learning rate schedule as implemented in Keras
    'Source: <https://keras.io/api/optimizers/learning_rate_schedules/exponential_decay/>'_  

    Parameters
    ----------
    eta0 : float (default=0.1)
        The initial learning rate that is decayed during optimization. This 
        will override the eta0 parameter on the estimator.

    eta_min : float(default=1e-4)
        The minimum allowable learning rate        

     decay_factor : float (default= 0.96)
        The factor used as the base of the polynomial used to compute
        the decayed learning rate.

    decay_steps : int
        The number of steps between each update

    staircase : Boolean (default=False)
        If true, the exponent is an integer division and decayed learning
        rate follows a staircase function. 
    """

    def __init__(self, eta0=0.1, eta_min=1e-4,
                       decay_factor=0.96, decay_steps=100, staircase=False):   
        super(ExponentialStepDecay, self).__init__(
            eta0=eta0,
            eta_min=eta_min,            
            decay_factor=decay_factor)    
        self.name = "Exponential Schedule Learning Rate Schedule"
        self.decay_steps = decay_steps
        self.staircase = staircase

    def _validate(self):
        """Performs hyperparameter """
        super(ExponentialStepDecay, self)._validate()
        validate_int(param=self.decay_steps, 
                     param_name='decay_steps',
                     minimum=1, left='open', right='open')        

    def _compute_learning_rate(self, epoch, log):
        exponent =  epoch // self._decay_step_length if self.staircase else \
            epoch / self._decay_step_length
        return self.eta0 * np.power(self.decay_factor, \
            exponent)

    def on_train_begin(self, log=None):
        super(ExponentialStepDecay, self).on_train_begin(log=log)
        self.decay_steps = min(self.decay_steps, self.model.epochs)
        self._decay_step_length = self.model.epochs // self.decay_steps        
        
# --------------------------------------------------------------------------  #
class PolynomialDecay(LearningRateSchedule):
    """ Polynomial decay learning rate schedule given by:
    .. math:: \eta_t=\eta_0 \cdot (1 - \frac{t}{\text{epochs}})^p

    Parameters
    ----------
    eta0 : float (default=0.1)
        The initial learning rate that is decayed during optimization. This 
        will override the eta0 parameter on the estimator.

    eta_min : float(default=1e-4)
        The minimum allowable learning rate        

    power : float (default=1)
        The power to which the polynomial is decayed 
    """

    def __init__(self, eta0=0.1, 
                       eta_min=1e-4,
                       power=1.0):        
        super(PolynomialDecay, self).__init__(
            eta0=eta0,
            eta_min=eta_min)   

        self.name = "Polynomial Decay Learning Rate Schedule"
        self.power = power

    def _validate(self):
        """Performs hyperparameter validation"""
        super(PolynomialDecay, self)._validate()
        validate_zero_to_one(param=self.power, 
                             param_name='power',
                             left='open', right='closed')                                

    def _compute_learning_rate(self, epoch, log):
        decay = (1 - (epoch / float(self.model.epochs))) ** self.power                
        return self.eta0 * decay

# --------------------------------------------------------------------------  #
class PolynomialStepDecay(LearningRateSchedule):
    """ Polynomial step decay learning rate as implemented in Keras and Tensorflow
    'Source: <https://keras.io/api/optimizers/learning_rate_schedules/polynomial_decay/>'_    

    Parameters
    ----------
    eta0 : float (default=0.1)
        The initial learning rate that is decayed during optimization. This 
        will override the eta0 parameter on the estimator.

    eta_min : float(default=1e-4)
        The minimum allowable learning rate        

    decay_steps : int (default=100)
        The total number of steps to decay the rate from the initial
        learning rate to the minimum learning rate.

    power : float (default=1)
        The power to which the polynomial is decayed 
    """

    def __init__(self, eta0=0.1, 
                       eta_min=1e-4,
                       decay_steps=100,
                       power=1.0):        
        super(PolynomialStepDecay, self).__init__(
            eta0=eta0,
            eta_min=eta_min)   

        self.name = "Polynomial Decay Learning Rate Schedule"
        self.decay_steps = decay_steps
        self.power = power

    def _validate(self):
        """Performs hyperparameter validation"""
        super(PolynomialStepDecay, self)._validate()
        
        validate_zero_to_one(param=self.power, 
                             param_name='power',
                             left='open', right='closed')    
        
        validate_int(param=self.decay_steps, 
                     param_name='decay_steps',
                     minimum=1, left='open', right='open')                

    def _compute_learning_rate(self, epoch, log):
        step = min(epoch, self.decay_steps)
        return (self.eta0 - self.eta_min) *\
            (1 - step / self.decay_steps) ** self.power +\
                self.eta_min

    def on_train_begin(self, log=None):
        super(PolynomialStepDecay, self).on_train_begin(log=log)
        self.decay_steps = min(self.decay_steps, self.model.epochs)
        

# --------------------------------------------------------------------------  #
class PowerSchedule(LearningRateSchedule):
    """ Exponential decay learning rate schedule as:
    .. math:: \eta_t=\eta_0 / (1+\frac{t}{r})^{c}

    Parameters
    ----------
    eta0 : float (default=0.1)
        The initial learning rate that is decayed during optimization. This 
        will override the eta0 parameter on the estimator.

    eta_min : float(default=1e-4)
        The minimum allowable learning rate        

    decay_steps : int (default=100)
        The total number of steps to decay the rate from the initial
        learning rate to the minimum learning rate.

    power : float (default=1)
        The power to which the polynomial is decayed 
    """
    def __init__(self, eta0=0.1, 
                       eta_min=1e-4,
                       decay_steps=100,
                       power=1.0):            
        super(PowerSchedule, self).__init__(
            eta0=eta0,
            eta_min=eta_min)
        self.name = "Power Schedule Learning Rate Schedule"              
        self.decay_steps = decay_steps
        self.power = power        

    def _validate(self):
        """Performs hyperparameter """
        super(PowerSchedule, self)._validate()
        validate_zero_to_one(param=self.power, 
                             param_name='power',
                             left='open', right='closed')    
        
        validate_int(param=self.decay_steps, 
                     param_name='decay_steps',
                     minimum=1, left='open', right='open') 

    def _compute_learning_rate(self, epoch, log):
        return self.eta0 / (1 + epoch/self.decay_steps)**self.power

# --------------------------------------------------------------------------  #
class BottouSchedule(LearningRateSchedule):
    """ Learning rate schedule as described in:
    https://cilvr.cs.nyu.edu/diglib/lsml/bottou-sgd-tricks-2012.pdf
    
    Parameters
    ----------
    eta0 : float (default=0.1)
        The initial learning rate that is decayed during optimization. This 
        will override the eta0 parameter on the estimator.

    eta_min : float(default=1e-4)
        The minimum allowable learning rate        

    decay_factor : float (default=0.5)
        The factor by which the learning rate is decayed.
    
    """

    def __init__(self, eta0=0.1, 
                       eta_min=1e-4,
                       decay_factor=0.5):    
        super(BottouSchedule, self).__init__(
            eta0=eta0,
            eta_min=eta_min,
            decay_factor=decay_factor)
        
        self.name = "Bottou Schedule Learning Rate Schedule"

    def _compute_learning_rate(self, epoch, log):
        return self.eta0 * (1 + self.eta0 * \
            self.decay_factor * epoch)**(-1)

# --------------------------------------------------------------------------- #
#                             IMPROVEMENT                                     #
# --------------------------------------------------------------------------- #
class Improvement(LearningRateSchedule):
    """Decays the learning rate if performance plateaus.

    Parameters
    ----------
    eta0 : float (default=0.1)
        The initial learning rate that is decayed during optimization. This 
        will override the eta0 parameter on the estimator.

    eta_min : float(default=1e-4)
        The minimum allowable learning rate

    decay_factor : float (default=0.5)
        The factor by which the learning rate is decayed when stabilization
        is encountered.

    metric : str, optional (default='train_cost')
        Specifies which statistic to metric for evaluation purposes.
        'train_cost': Training set costs
        'train_score': Training set scores based upon the model's metric parameter
        'val_cost': Validation set costs
        'val_score': Validation set scores based upon the model's metric parameter
        'gradient_norm': The norm of the gradient of the objective function w.r.t. theta

    epsilon : float, optional (default=0.001)
        The factor by which performance is considered to have improved. For 
        instance, a value of 0.01 means that performance must have improved
        by a factor of 1% to be considered an improvement.

    patience : int, optional (default=10)
        The number of consecutive epochs of non-improvement that would 
        stop training.    
    """

    def __init__(self, eta0=0.1, eta_min=1e-4,
                 decay_factor=0.5, metric='train_cost',  epsilon=0.001, 
                 patience=10):
        super(Improvement, self).__init__(
            eta0=eta0,
            eta_min=eta_min,
            decay_factor=decay_factor)

        self.name = "Improvement Learning Rate Schedule"        
        self.metric = metric                
        self.epsilon = epsilon
        self.patience = patience        

    def on_train_begin(self, log=None):        
        """Sets key variables at beginning of training.
        
        Parameters
        ----------
        log : dict
            Contains no information
        """
        super(Improvement, self).on_train_begin(log)
        self._observer = PerformanceObserver(metric=self.metric, \
            epsilon=self.epsilon, patience=self.patience)    
        self._observer.on_train_begin()        

    def _compute_learning_rate(self, epoch, log):
        lr = log.get('eta') * self.decay_factor \
            if self._observer.stabilized else log.get('eta')
        return lr


    def on_epoch_end(self, epoch, log=None):
        """Determines whether convergence has been achieved.
        Parameters
        ----------
        epoch : int
            The current epoch number
        log : dict
            Dictionary containing training cost, (and if metric=score, 
            validation cost)  
        """        
        log = log or {}                        
        self._observer.on_epoch_end(epoch, log)
        super(Improvement, self).on_epoch_end(epoch, log)        
        
           


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : task.py                                                           #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Tuesday, May 19th 2020, 10:00:13 pm                         #
# Last Modified : Tuesday, May 19th 2020, 10:00:13 pm                         #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Defines linear, logistic, and multinomial logistic regression classes."""
from abc import ABC, abstractmethod 

import numpy as np
from sklearn.base import BaseEstimator

from mlstudio.supervised.core.activations import Sigmoid, Softmax
from mlstudio.supervised.core.objectives import MSE, CrossEntropy, CategoricalCrossEntropy 
# --------------------------------------------------------------------------  #
class Task(ABC, BaseEstimator):
    """Defines the base class for all tasks."""

    @abstractmethod
    def name(self):
        pass

    @abstractmethod
    def compute_output(self, theta, X):
        """Computes output for the task."""
        pass

    @abstractmethod
    def predict(self, theta, X):
        """Computes prediction."""
        pass

# --------------------------------------------------------------------------  #
class LinearRegression(Task):
    """Defines the linear regression task."""

    @property
    def name(self):
        return "Linear Regression"
    
    def compute_output(self, theta, X):
        """Computes linear regression output."""        
        return X.dot(theta) 

    def predict(self, theta, X):
        return self.compute_output(theta, X)

# --------------------------------------------------------------------------  #
class LogisticRegression(Task):
    """Defines the logistic regression task."""

    def __init__(self):
        self.sigmoid = Sigmoid()

    @property
    def name(self):
        return "Logistic Regression"    

    def compute_output(self, theta, X):
        """Computes logistic regression output."""        
        z = X.dot(theta)
        return self.sigmoid(z)

    def predict(self, theta, X):
        o = self.compute_output(theta, X)        
        return np.round(o).astype(int)


# --------------------------------------------------------------------------  #
class MultinomialLogisticRegression(Task):
    """Defines the multinomial logistic regression task."""

    def __init__(self):
        self.softmax = Softmax()

    @property
    def name(self):
        return "Multinomial Logistic Regression"    

    def compute_output(self, theta, X):
        """Computes multinomial logistic regression output."""        
        z = X.dot(theta)
        return self.softmax(z)        

    def predict(self, theta, X):
        o = self.compute_output(theta, X)        
        return o.argmax(axis=1)


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.14                                                            #
# File    : test_objectives.py                                                #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Monday, June 15th 2020, 3:45:31 pm                          #
# Last Modified : Monday, June 15th 2020, 3:45:31 pm                          #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
import math
import numpy as np
import pytest
from pytest import mark

from mlstudio.supervised.core.regularizers import L1, L2, L1_L2
# --------------------------------------------------------------------------  #
@mark.regularizers
@mark.lasso
class LassoTests:

    def test_lasso(self, get_regularization_package):
        regularization_package = get_regularization_package
        theta = np.array([40, 6, -22, 31,26])   
        m = len(theta)
        alpha = 0.1                
        lasso = L1(alpha=alpha)
        # Lasso loss regularization        
        exp_result = regularization_package['l1_cost'] 
        act_result = lasso(theta, m)
        assert np.isclose(exp_result,act_result), "Lasso regularization cost error"

    def test_lasso_gradient(self, get_regularization_package):
        regularization_package = get_regularization_package
        theta = np.array([40, 6, -22, 31,26])   
        m = len(theta)
        alpha = 0.1                
        lasso = L1(alpha=alpha)
        # Lasso loss regularization        
        exp_result = regularization_package['l1_grad'] 
        act_result = lasso.gradient(theta, m)
        assert np.allclose(exp_result,act_result), "Lasso regularization gradient error"

# --------------------------------------------------------------------------  #
@mark.regularizers
@mark.ridge
class RidgeTests:

    def test_ridge(self, get_regularization_package):
        regularization_package = get_regularization_package
        theta = np.array([40, 6, -22, 31,26])   
        m = len(theta)
        alpha = 0.1                
        ridge = L2(alpha=alpha)
        # Lasso loss regularization        
        exp_result = regularization_package['l2_cost'] 
        act_result = ridge(theta, m)
        assert np.isclose(exp_result,act_result), "Ridge regularization cost error"

    def test_ridge_gradient(self, get_regularization_package):
        regularization_package = get_regularization_package
        theta = np.array([40, 6, -22, 31,26])   
        m = len(theta)
        alpha = 0.1                
        ridge = L2(alpha=alpha)
        # Lasso loss regularization        
        exp_result = regularization_package['l2_grad'] 
        act_result = ridge.gradient(theta, m)
        assert np.allclose(exp_result,act_result), "Ridge regularization gradient error"

# --------------------------------------------------------------------------  #
@mark.regularizers
@mark.elasticnet
class ElasticNetTests:

    def test_elasticnet(self, get_regularization_package):
        regularization_package = get_regularization_package
        theta = np.array([40, 6, -22, 31,26])   
        m = len(theta)
        alpha = 0.1            
        ratio = 0.5    
        elasticnet = L1_L2(alpha=alpha, ratio=ratio)
        # Lasso loss regularization        
        exp_result = regularization_package['l1_l2_cost'] 
        act_result = elasticnet(theta, m)
        assert np.isclose(exp_result,act_result), "ElasticNet regularization cost error"

    def test_elasticnet_gradient(self, get_regularization_package):
        regularization_package = get_regularization_package
        theta = np.array([40, 6, -22, 31,26])   
        m = len(theta)
        alpha = 0.1
        ratio = 0.5                
        elasticnet = L1_L2(alpha=alpha, ratio=ratio)
        # Lasso loss regularization        
        exp_result = regularization_package['l1_l2_grad'] 
        act_result = elasticnet.gradient(theta, m)
        assert np.allclose(exp_result,act_result), "ElasticNet regularization gradient error"



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.14                                                            #
# File    : test_activations.py                                               #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Monday, June 15th 2020, 10:24:08 pm                         #
# Last Modified : Monday, June 15th 2020, 10:24:25 pm                         #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Test Activation Functions."""
import math
import numpy as np
import pytest
from pytest import mark

from scipy.special import expit
from sklearn.utils.extmath import softmax

from mlstudio.supervised.core.activations import Sigmoid, Softmax
# --------------------------------------------------------------------------  #
@mark.activations
@mark.sigmoid
class SigmoidActivationTests:

    def test_sigmoid_activation(self):
        # Get an array
        x = np.random.default_rng().uniform(low=0, high=20, size=20)
        # Expected result
        exp_result = expit(x)
        # Actual result
        act = Sigmoid()
        act_result = act(x)
        assert np.allclose(exp_result, act_result), "Sigmoid error"        

@mark.activations
@mark.softmax
class SoftmaxActivationTests:

    def test_softmax_activation(self):
        # Get an array
        x = np.random.default_rng().uniform(low=0, high=20, size=(20,4))
        # Expected result        
        exp_result = softmax(x)
        # Actual result
        act = Softmax()
        act_result = act(x)
        assert np.allclose(exp_result, act_result), "Softmax error"        








#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : gradient_descent_optimizers.py                                    #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Saturday, May 16th 2020, 9:13:15 pm                         #
# Last Modified : Saturday, May 16th 2020, 9:13:16 pm                         #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Gradient descent optimization algorithms."""
from abc import ABC, abstractmethod
import math

import numpy as np
from sklearn.base import BaseEstimator  
# --------------------------------------------------------------------------  #
class Optimizer(ABC, BaseEstimator):
    """Base class for all optimizers."""

    def __init__(self):
        pass

    @abstractmethod        
    def __call__(self, gradient, learning_rate, theta, **kwargs):   
        """Computes the parameter updates.
        
        Parameters
        ----------
        gradient : func
            The function that performs the gradient computation 

        learning_rate : float
            The learning rate from the estimator object.

        theta : array-like
            The model parameters

        **kwargs : dict
            Arbitrary parameters used for computing the gradient.

        Returns
        -------
        theta : array-like
            The updated parameters of the model

        grad : array-like
            The gradient of the objective function w.r.t. parameters theta.
        """
        pass



# --------------------------------------------------------------------------  #
class GradientDescentOptimizer(Optimizer):
    """Standard gradient descent optimizer."""

    def __init__(self):
        self.name = "Gradient Descent"
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):        
        grad = gradient(theta, **kwargs)
        theta = theta - learning_rate * grad
        return theta, grad

# --------------------------------------------------------------------------  #
class Momentum(Optimizer):
    """Standard gradient descent optimizer."""

    def __init__(self, gamma=0.9):
        self.name = "Momentum"
        self.gamma = gamma
        self._velocity = 0
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):             
        grad = gradient(theta, **kwargs)
        self._velocity = self.gamma * self._velocity + learning_rate * grad
        theta = theta - self._velocity
        return theta, grad

# --------------------------------------------------------------------------  #
class Nesterov(Optimizer):
    """Nesterov accelerated gradient optimizer."""

    def __init__(self, gamma=0.9):
        self.name = "Nesterov"
        self.gamma = gamma
        self._velocity = 0
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):
        next_position = theta - self.gamma * self._velocity        
        grad = gradient(next_position, **kwargs)
        self._velocity = self.gamma * self._velocity + learning_rate * grad
        theta = theta - self._velocity
        return theta, grad

# --------------------------------------------------------------------------  #
class Adagrad(Optimizer):
    """Adagrad optimizer."""

    def __init__(self, epsilon=1e-8):
        self.name = "Adagrad"
        self.epsilon = epsilon
        self.gradients = 0
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):
        grad = gradient(theta, **kwargs)
        # Create effective learning rate
        self.gradients = self.gradients + np.square(grad)     
        self.gradients = self.gradients + np.array(self.epsilon, ndmin=1)        
        elr = learning_rate / np.sqrt(np.array(self.gradients, ndmin=1))        
        # Convert to diagonal matrix
        Gt = np.diag(elr)
        theta = theta - Gt.dot(grad)        
        return theta, grad        

# --------------------------------------------------------------------------  #
class Adadelta(Optimizer):
    """Adadelta optimizer."""

    def __init__(self, gamma=0.9, epsilon=1e-8):
        self.name = "Adadelta"
        self.gamma = gamma
        self.epsilon = epsilon
        self.avg_sqr_gradient = 0
        self.avg_sqr_delta_theta = 0
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):                
        grad = gradient(theta, **kwargs)     

        self.avg_sqr_gradient = self.gamma * self.avg_sqr_gradient + \
            (1 - self.gamma) * np.square(grad)            
        rms_grad = np.sqrt(np.array(np.square(np.array(self.avg_sqr_gradient, ndmin=1)) + np.array(self.epsilon, ndmin=1), ndmin=1))
        
        delta_theta = -learning_rate / rms_grad  * grad            

        self.avg_sqr_delta_theta = self.gamma * self.avg_sqr_delta_theta + \
            (1 - self.gamma) * np.square(np.array(delta_theta, ndmin=1))
        rms_delta_theta = np.sqrt(np.array(np.square(np.array(self.avg_sqr_delta_theta, ndmin=1)) + np.array(self.epsilon, ndmin=1),ndmin=1))

        delta_theta = - (rms_delta_theta / rms_grad).dot(grad)

        theta = theta + delta_theta
        return theta, grad

# --------------------------------------------------------------------------  #
class RMSprop(Optimizer):
    """RMSprop optimizer."""

    def __init__(self, gamma=0.9, epsilon=1e-8):     
        self.name = "RMSprop"   
        self.gamma = gamma
        self.epsilon = epsilon
        self.avg_sqr_gradient = 0
        self.avg_sqr_delta_theta = 0
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):                
        grad = gradient(theta, **kwargs)        
        self.avg_sqr_gradient = self.gamma * self.avg_sqr_gradient + \
            0.1 * np.square(grad)
        rms_grad = np.sqrt(np.array(np.square(np.array(self.avg_sqr_gradient, ndmin=1)) + np.array(self.epsilon, ndmin=1),ndmin=1))
        theta = theta - (learning_rate / rms_grad) * grad
        
        return theta, grad

# --------------------------------------------------------------------------  #
class Adam(Optimizer):
    """Adam optimizer."""

    def __init__(self, beta_one=0.9, beta_two=0.999, epsilon=10e-8):
        self.name = "Adam"
        self.beta_one = beta_one
        self.beta_two = beta_two        
        self.epsilon = epsilon

        self.t = 0
        self.m = 0        
        self.v = 0
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):                
        self.t += 1
        grad = gradient(theta, **kwargs)        
        self.m = self.beta_one * self.m + (1 - self.beta_one) * grad
        self.v = self.beta_two * self.v + (1 - self.beta_two) * np.square(grad)
        # Bias corrected moment estimates
        m_hat = self.m / (1 - self.beta_one**self.t)
        v_hat = self.v / (1 - self.beta_two**self.t)

        theta = theta - learning_rate / (np.sqrt(np.array(v_hat)) + np.array(self.epsilon)) * m_hat        
        
        return theta, grad        

# --------------------------------------------------------------------------  #
class AdaMax(Optimizer):
    """AdaMax optimizer."""

    def __init__(self, beta_one=0.9, beta_two=0.999):    
        self.name = "AdaMax"   
        self.beta_one = beta_one
        self.beta_two = beta_two        
        self.t = 0
        self.m = 0
        self.u = 0
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):                
        self.t += 1
        grad = gradient(theta, **kwargs)                
        self.m = self.beta_one * self.m + (1 - self.beta_one) * grad
        m_hat = self.m / (1 - self.beta_one**self.t)
        self.u = np.maximum(self.beta_two * self.u, np.linalg.norm(grad,1))    
        theta = theta - (learning_rate / (1-self.beta_one**self.t)) * m_hat / self.u
        
        return theta, grad              

# --------------------------------------------------------------------------  #
class Nadam(Optimizer):
    """Nadam optimizer."""

    def __init__(self, beta_one=0.9, beta_two=0.999, epsilon=10e-8):
        self.name = "Nadam"
        self.beta_one = beta_one        
        self.beta_two = beta_two        
        self.epsilon = epsilon
        self.t = 0
        self.m = 0
        self.v = 0
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):    
        self.t += 1
        grad = gradient(theta, **kwargs)        
        self.m = self.beta_one * self.m + (1 - self.beta_one) * grad   
        self.v = self.beta_two * self.v + (1 - self.beta_two) * np.square(grad)
        # Bias corrected moment estimates
        m_hat = self.m / (1 - self.beta_one**self.t)
        v_hat = self.v / (1 - self.beta_two**self.t)
     
        # Nadam update
        theta = theta - (learning_rate / (np.sqrt(np.array(v_hat,ndmin=1)) + self.epsilon)) * \
            (self.beta_one * m_hat + ((1-self.beta_one)* grad)/(1-self.beta_one**self.t))

        return theta, grad                                    

# --------------------------------------------------------------------------  #
class AMSGrad(Optimizer):
    """AMSGrad optimizer."""

    def __init__(self, beta_one=0.9, beta_two=0.999, epsilon=10e-8):
        self.name = "AMSGrad"
        self.beta_one = beta_one        
        self.beta_two = beta_two
        self.epsilon = epsilon
        self.t = 0
        self.m = 0
        self.v = 0
        self.v_hat = 0
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):    
        self.t += 1
        grad = gradient(theta, **kwargs)        
        self.m = self.beta_one * self.m + (1 - self.beta_one) * grad
        self.v = self.beta_two * self.v + (1 - self.beta_two) * np.square(grad)
        self.v_hat = np.array(np.maximum(self.v_hat, self.v),ndmin=1)
        theta = theta - (learning_rate / (np.sqrt(np.array(self.v_hat)) + np.array(self.epsilon))) * self.m
        
        return theta, grad                        

# --------------------------------------------------------------------------  #
class AdamW(Optimizer):
    """AdamW optimizer."""

    def __init__(self, beta_one=0.9, beta_two=0.999, decay_rate=1e-4, epsilon=10e-8):
        self.name = "AdamW"
        self.beta_one = beta_one        
        self.beta_two = beta_two
        self.decay_rate = decay_rate
        self.epsilon = epsilon
        self.t = 0
        self.m = 0
        self.v = 0
        self.v_hat = 0
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):    
        self.t += 1
        grad = gradient(theta, **kwargs)        
        self.m = self.beta_one * self.m + (1 - self.beta_one) * grad
        self.v = self.beta_two * self.v + (1 - self.beta_two) * np.square(grad)
        self.v_hat = np.array(np.maximum(self.v_hat, self.v))
        theta = theta - (learning_rate / (np.sqrt(np.array(self.v_hat,ndmin=1)) + np.array(self.epsilon, ndmin=1))) * \
            self.m + self.decay_rate * theta
        
        return theta, grad          

# --------------------------------------------------------------------------  #
class QHAdam(Optimizer):
    """Quasi-Hyperbolic Adam"""

    def __init__(self, beta_one=0.9, beta_two=0.999, decay_rate=1e-4, epsilon=10e-8,
                 v1=1, v2=1):
        self.name = "QH-Adam"
        self.v1 = v1
        self.v2 = v2
        self.beta_one = beta_one        
        self.beta_two = beta_two
        self.epsilon = epsilon
        self.t = 0
        self.m = 0
        self.v = 0
        self.v_hat = 0
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):    
        self.t += 1
        grad = gradient(theta, **kwargs)        
        self.m = self.beta_one * self.m + (1 - self.beta_one) * grad
        self.v = self.beta_two * self.v + (1 - self.beta_two) * np.square(grad)
        self.v_hat = np.maximum(self.v_hat, self.v)
        theta = theta - learning_rate * (((1-self.v1)*grad + self.v1 * self.m) /\
            (np.sqrt(np.array(1-self.v2,ndmin=1))*np.square(np.array(grad,ndmin=1))+self.v2*self.v+self.epsilon))
        
        return theta, grad                  

# --------------------------------------------------------------------------  #
class AggMo(Optimizer):
    """Aggregated Momentum."""

    def __init__(self, k=3, betas=[0, 0.9, 0.999], decay_factor=0, epsilon=10e-8):
        self.name = "AggMo"
        self.k = k
        self.betas = betas
        self.t = 0
        self.v = 0
        
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):    
        self.t += 1
        grad = gradient(theta, **kwargs)       
        if self.v == 0:
            self.v = {}            
            for beta in self.betas:
                self.v[beta] = np.zeros_like(theta)
        buf = np.zeros_like(theta)
        for beta in self.betas:
            self.v[beta] = beta * self.v[beta] - grad
            buf = buf + self.v[beta]
        buf = buf / self.k

        theta = theta - learning_rate * buf 
        
        return theta, grad        

# --------------------------------------------------------------------------  #
class QuasiHyperbolicMomentum(Optimizer):
    """Quasi-Hyperbolic Momentum"""

    def __init__(self, v=0.7, beta=0.999, epsilon=10e-8):
        self.name = "QH-Momentum"       
        self.v = v 
        self.beta = beta
        self.t = 0
        self.m = 0
        
    
    def __call__(self, gradient, learning_rate, theta, **kwargs):    
        self.t += 1
        grad = gradient(theta, **kwargs)
        self.m = self.beta * self.m + (1 - self.beta) * grad
        theta = theta - learning_rate * ((1-self.v) * grad + self.v * self.m)
        
        return theta, grad                 

         


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.14                                                            #
# File    : test_performance_observer.py                                      #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Monday, June 15th 2020, 8:59:49 am                          #
# Last Modified : Monday, June 15th 2020, 8:59:49 am                          #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Tests observer."""
from collections import OrderedDict
import os
import math
import os
from pathlib import Path
import sys
testdir = str(Path(__file__).parents[2])
testdatadir = os.path.join(testdir, 'test_data')
sys.path.append(testdatadir)

import numpy as np
import pandas as pd
import pytest
from pytest import mark
from tabulate import tabulate

from mlstudio.supervised.observers.early_stop import Performance

# --------------------------------------------------------------------------  #
#                          TEST OBSERVER                                      #
# --------------------------------------------------------------------------  #
@mark.observer
@mark.performance
class PerformanceTests:
    def test_performance_validation(self): 
        # Validate metric  
        with pytest.raises(ValueError) as v:
            observer = Performance(metric='hair')
            observer.on_train_begin()
        with pytest.raises(TypeError) as v:
            observer = Performance(metric=1)
            observer.on_train_begin()
        # Validate epsilon
        with pytest.raises(TypeError) as v:
            observer = Performance(epsilon='hair')                
            observer.on_train_begin()        
        with pytest.raises(ValueError) as v:
            observer = Performance(epsilon=1.1)                
            observer.on_train_begin()
        with pytest.raises(ValueError) as v:
            observer = Performance(epsilon=-0.1)                
            observer.on_train_begin()        
        # Validate patience
        with pytest.raises(TypeError) as v:
            observer = Performance(patience='hair')        
            observer.on_train_begin()
        with pytest.raises(ValueError) as v:
            observer = Performance(patience=0)        
            observer.on_train_begin()        

    def _get_expected_results(self, filepath):    
        return pd.read_excel(filepath, sheet_name='results', header=0, 
                                names=['epoch', 'train_cost', 'baseline', 'improvement',
                                    'sig', 'iter_no_change','factor', 'stable', 'best'], 
                                usecols="A:I",
                                dtype={"sig":bool,"stable":bool})
            

    @mark.observer
    @mark.performance
    def test_performance_reporting(self, get_mock_estimator):    
        # Obtain expected results
        filepath = os.path.join(testdatadir, "test_performance_observer_cost.xlsx")
        exp_results = self._get_expected_results(filepath)    

        # Create dict of logs dicts
        logs = exp_results.to_dict('index')
        
        # Create an estimator
        Estimator = get_mock_estimator
        est = Estimator()
        # Obtain and on_train_begin observer
        observer = Performance(mode='passive', metric='train_cost', 
                               epsilon=0.02, patience=5)
        observer.on_train_begin()

        # Register with the observer
        observer.set_model(est)
        
        # Iterate through logs
        for epoch, log in logs.items():
            observer.on_epoch_end(epoch, log)

        # End training
        observer.on_train_end()
            
        # Gather expected results in numpy format
        exp_improvement = exp_results['sig'].to_numpy()
        exp_stability = exp_results['stable'].to_numpy()
        

        # Gather actual results and convert to numpy format
        log = observer.performance_log_
        act_improvement = log['significant_improvement']
        act_stability = log['stabilized']

        # Compare expected and actual results
        assert np.array_equal(exp_improvement, act_improvement), "Improvement errors: \nExp {e} \nAct {a}".format(\
            e=str(exp_improvement), a=str(act_improvement))
        assert np.array_equal(exp_stability, act_stability), "Stability errors"
        
        print(tabulate(log, headers="keys"))


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.14                                                            #
# File    : test_optimizer.py                                                 #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Tuesday, June 16th 2020, 12:02:30 am                        #
# Last Modified : Tuesday, June 16th 2020, 12:02:31 am                        #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Integration test for gradient descent pure optimizer class."""
import math
import numpy as np
import pytest
from pytest import mark

from mlstudio.supervised.machine_learning.gradient_descent import GradientDescentPureOptimizer
from mlstudio.supervised.observers.learning_rate import TimeDecay, StepDecay
from mlstudio.supervised.observers.learning_rate import ExponentialDecay
from mlstudio.supervised.observers.learning_rate import ExponentialStepDecay
from mlstudio.supervised.observers.learning_rate import PolynomialDecay
from mlstudio.supervised.observers.learning_rate import PolynomialStepDecay
from mlstudio.supervised.observers.learning_rate import PowerSchedule
from mlstudio.supervised.observers.learning_rate import BottouSchedule
from mlstudio.supervised.observers.learning_rate import Improvement
from mlstudio.supervised.observers.early_stop import Performance
from mlstudio.supervised.core.optimizers import GradientDescentOptimizer
from mlstudio.supervised.core.optimizers import Momentum
from mlstudio.supervised.core.optimizers import Nesterov
from mlstudio.supervised.core.optimizers import Adagrad
from mlstudio.supervised.core.optimizers import Adadelta
from mlstudio.supervised.core.optimizers import RMSprop
from mlstudio.supervised.core.optimizers import Adam, AdaMax, Nadam
from mlstudio.supervised.core.optimizers import AMSGrad, AdamW, QHAdam
from mlstudio.supervised.core.optimizers import QuasiHyperbolicMomentum
from mlstudio.supervised.core.objectives import Adjiman, BartelsConn
from mlstudio.supervised.core.objectives import Himmelblau, Leon
from mlstudio.supervised.core.objectives import Rosenbrock, Branin02
from mlstudio.supervised.core.objectives import StyblinskiTank
from mlstudio.supervised.core.objectives import ThreeHumpCamel, Ursem01
from mlstudio.supervised.core.objectives import Wikipedia
# --------------------------------------------------------------------------  #
#                       TEST OBJECTIVE FUNCTIONS                              #
# --------------------------------------------------------------------------  #
@mark.gradient_descent
@mark.pure_optimizer
class PureOptimizerObjectiveTests:

    objectives = [Adjiman(), BartelsConn(), Himmelblau(), Leon(), Rosenbrock(), Branin02(),
                  StyblinskiTank(), ThreeHumpCamel(), Ursem01(), Wikipedia()]

    optimizers = [GradientDescentOptimizer(), Momentum(), Nesterov(),
                  Adagrad(), Adadelta(), RMSprop(), Adam(), AdaMax(), Nadam(),
                  AMSGrad(), AdamW(), QHAdam(), QuasiHyperbolicMomentum()]

    def test_pure_optimizer_core(self):
        epochs = 500
        for objective in self.objectives:
            objective_min_norm = np.linalg.norm(objective.minimum)
            for optimizer in self.optimizers:
                est = GradientDescentPureOptimizer(epochs=epochs, 
                                                   optimizer=optimizer,
                                                   objective=objective)
                est.fit()
                bb = est.blackbox_
                solution_norm = np.linalg.norm(bb.epoch_log.get('theta')[-1])
                assert len(bb.epoch_log.get('epoch')) == epochs, "Epoch log wrong length"
                assert len(bb.epoch_log.get('theta')) == epochs, "Epoch log wrong length"
                assert len(bb.epoch_log.get('train_cost')) == epochs, "Epoch log wrong length"
                assert len(bb.epoch_log.get('eta')) == epochs, "Epoch log wrong length"
                msg = "\nPoor solution for objective = {o}, optimizer = {p}\n       min_norm = {m}, solution_norm = {s}".format(o = objective.__class__.__name__,
                                                                                  p = optimizer.__class__.__name__,
                                                                                  m = str(objective_min_norm),
                                                                                  s = str(solution_norm))
                if solution_norm - objective_min_norm > 50:
                    print(msg)

    


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : MLStudio                                                          #
# Version : 0.1.0                                                             #
# File    : file_manager.py                                                   #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Friday, April 10th 2020, 8:02:25 pm                         #
# Last Modified : Friday, April 10th 2020, 8:02:25 pm                         #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
import os
import time

from matplotlib import animation, rc
from matplotlib.animation import FuncAnimation
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.offline as py

from  mlstudio.utils.format import snake

def check_directory(directory):    
    if not os.path.exists(directory):
        os.makedirs(directory)               

def save_df(df, directory, filename, mode='a'):
    if not os.path.exists(directory):
        os.makedirs(directory)
    
    path = os.path.join(os.path.abspath(directory), filename)
    
    if os.path.exists(path) and mode == 'a':
        df.to_csv(path, mode=mode, header=False)
    else:
        df.to_csv(path)        

def save_fig(fig, directory, filename):
    if os.path.exists(directory):
        path = os.path.join(os.path.abspath(directory), filename)
        fig.savefig(path, facecolor='w', bbox_inches=None)
    else:
        os.makedirs(directory)
        path = os.path.join(os.path.abspath(directory),filename)
        fig.savefig(path, facecolor='w', bbox_inches=None)

def save_gif(ani, directory, filename, fps):
    face_edge_colors = {'facecolor': 'w', 'edgecolor': 'w'}
    path = os.path.join(os.path.abspath(directory), filename)
    if os.path.exists(directory):
        ani.save(path, writer='imagemagick', fps=fps, savefig_kwargs = face_edge_colors)
    else:
        os.makedirs(directory)                
        ani.save(path, writer='imagemagick', fps=fps, savefig_kwargs = face_edge_colors)

def save_csv(df, directory, filename):
    path = os.path.join(os.path.abspath(directory), filename)
    if os.path.exists(directory):
        df.to_csv(path, index=False)
    else:
        os.makedirs(directory)                
        df.to_csv(path, index=False)

def save_numpy(a, directory, filename):
    path = os.path.join(os.path.abspath(directory), filename)
    if os.path.exists(directory):
        np.save(file=path, arr=a)
    else:
        os.makedirs(directory)                
        np.save(file=path, arr=a)        

def save_plotly(a, directory, filename):
    path = os.path.join(os.path.abspath(directory), filename)
    if os.path.exists(directory):
        py.plot(a, filename=path, auto_open=False, include_mathjax='cdn')
    else:
        os.makedirs(directory)                
        py.plot(a, filename=path, auto_open=False, include_mathjax='cdn')

def get_filename(instance, fileext, element=None):
        """Creates a standard format filename for saving plots."""    

        # Obtain user id, class name and date time 
        project = "ml_studio_"        
        userhome = os.path.expanduser('~')          
        username = os.path.split(userhome)[-1] + "_"
        clsname = instance.__class__.__name__ + "_"
        if element:
            element = element + "_"
        else:
            element = ""
        timestr = time.strftime("%Y%m%d-%H%M%S")
        # Snake case format filename
        filename = project + username + clsname + element + timestr + fileext
        filename = snake(filename)        
        return filename


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.14                                                            #
# File    : history.py                                                        #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Friday, May 15th 2020, 9:16:41 pm                           #
# Last Modified : Sunday, June 14th 2020, 11:49:38 pm                         #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Module containing observers that monitor and report on optimization."""
import datetime
import numpy as np
import pandas as pd

from mlstudio.supervised.observers.base import Observer, PerformanceObserver
from mlstudio.utils.validation import validate_metric, validate_int
from mlstudio.utils.validation import validate_zero_to_one
# --------------------------------------------------------------------------- #
#                                BLACKBOX                                     #
# --------------------------------------------------------------------------- #
class BlackBox(Observer):
    """Repository for data obtained during optimization."""

    def __init__(self):
        super(BlackBox, self).__init__()
        self.name = "BlackBox"

    def on_train_begin(self, log=None):
        """Sets instance variables at the beginning of training.
        
        Parameters
        ----------
        log : Dict
            Dictionary containing the X and y data
        """ 
        self.total_epochs = 0
        self.total_batches = 0
        self.start = datetime.datetime.now()
        self.epoch_log = {}
        self.batch_log = {}        

    def on_train_end(self, log=None):        
        """Sets instance variables at end of training.
        
        Parameters
        ----------
        log : Dict
            Not used 
        """
        self.end = datetime.datetime.now()
        self.duration = (self.end-self.start).total_seconds() 
        final_results = {}
        for k, v in self.epoch_log.items():
            final_results[k] = v[-1]
        self.model.final_results_ = final_results

    def on_batch_end(self, batch, log=None):
        """Updates data and statistics relevant to the training batch.
        
        Parameters
        ----------
        batch : int
            The current training batch
        
        log : dict
            Dictionary containing batch statistics, such as batch size, current
            weights and training cost.
        """
        self.total_batches += 1
        for k,v in log.items():
            self.batch_log.setdefault(k,[]).append(v)        

    def on_epoch_begin(self, epoch, log=None):
        """Updates data and statistics relevant to the training epoch.

        Parameters
        ----------
        epoch : int
            The current training epoch
        
        log : dict
            Dictionary containing data and statistics for the current epoch,
            such as weights, costs, and optional validation set statistics
            beginning with 'val_'.
        """
        log = log or {}
        self.total_epochs += 1
        for k,v in log.items():
            self.epoch_log.setdefault(k,[]).append(v)

   

# --------------------------------------------------------------------------- #
#                                PROGRESS                                     #
# --------------------------------------------------------------------------- #              
class Progress(Observer):
    """Class that reports progress at designated points during training."""

    def __init__(self):
        super(Progress, self).__init__()
        self.name = "Progress"    
    
    def on_epoch_begin(self, epoch, log=None):
        """Reports progress at the end of each epoch.

        Parameters
        ----------
        epoch : int
            The current training epoch

        log : Dict
            Statistics obtained at end of epoch
        """
        if self.model.verbose:
            if not isinstance(self.model.verbose, int):
                raise TypeError("Verbose must be False or an integer. The \
                    integer indicates the number of epochs between each \
                    progress update.")            
            else:
                if epoch % self.model.verbose == 0:
                    items_to_report = ('epoch', 'train', 'val')
                    log = {k:v for k,v in log.items() if k.startswith(items_to_report)}
                    progress = "".join(str(key) + ': ' + str(np.round(value,4)) + ' ' \
                        for key, value in log.items())
                    print(progress)
        


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : benchmark.py                                                      #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Sunday, May 24th 2020, 11:06:10 am                          #
# Last Modified : Sunday, June 14th 2020, 9:48:14 pm                          #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
#%%
from collections import OrderedDict
from datetime import datetime
import os
from pathlib import Path
import sys

import pandas as pd
import numpy as np
homedir = str(Path(__file__).parents[3])
demodir = str(Path(__file__).parents[1])
sys.path.append(homedir)

from mlstudio.supervised.machine_learning.gradient_descent import GradientDescentPureOptimizer
from mlstudio.visual.animations import animate_optimization
from mlstudio.supervised.core.objectives import Adjiman, StyblinskiTank, Wikipedia
from mlstudio.supervised.core.objectives import ThreeHumpCamel, Ursem01, Branin02
from mlstudio.supervised.core.optimizers import GradientDescentOptimizer, Momentum, Nesterov
from mlstudio.supervised.core.optimizers import Adagrad, Adadelta, RMSprop
from mlstudio.supervised.core.optimizers import Adam, AdaMax, AdamW
from mlstudio.supervised.core.optimizers import Nadam, AMSGrad, QHAdam
from mlstudio.supervised.core.optimizers import QuasiHyperbolicMomentum
from mlstudio.supervised.core.optimizers import AggMo
from mlstudio.utils.data_analyzer import cosine
from mlstudio.utils.file_manager import save_df

# --------------------------------------------------------------------------  #
# Designate file locations
figures = os.path.join(demodir, "figures")
# --------------------------------------------------------------------------  #
# Package up the objective functions
optimizers = [Momentum(), Nesterov(), Adagrad(), Adadelta(), RMSprop(), Adam(),
              AdaMax(), Nadam(), AMSGrad(), AdamW(), QHAdam(),
              QuasiHyperbolicMomentum()]

objectives = [Adjiman(), Branin02(), Ursem01(),
              StyblinskiTank(), ThreeHumpCamel(), Wikipedia()]

# --------------------------------------------------------------------------  #
# Train models
solutions = OrderedDict()
results = []
for objective in objectives:
    estimators = OrderedDict()   

    for optimizer in optimizers:
        estimators[optimizer.name] = {}
        model = GradientDescentPureOptimizer(learning_rate=0.01,
                                theta_init=objective.start, 
                                epochs=500, objective=objective,
                                optimizer=optimizer)
        model.fit()
        sim = cosine(objective.minimum, model.theta_)
        d = {}
        d['DateTime'] = datetime.now().strftime("%Y/%m/%d %H:%M:%S")
        d['Objective'] = model.objective.name
        d['Optimizer'] = optimizer.name
        d['Epochs'] = model.epochs        
        d['Starting Learning Rate'] = model.learning_rate
        d['Final Learning Rate'] = model.eta
        if model.schedule:
            d['Schedule'] = model.schedule.name              
        else:
            d['Schedule'] = None            
        d['gradient_min'] = np.min(model.blackbox_.epoch_log.get('gradient_norm'))
        d['gradient_max'] = np.max(model.blackbox_.epoch_log.get('gradient_norm'))        
        d['gradient_mean'] = np.mean(model.blackbox_.epoch_log.get('gradient_norm'))        
        d['True'] = objective.minimum
        d['est'] = model.theta_
        d['sim'] = sim
        results.append(d)
        estimators[optimizer.name]['model'] = model
        estimators[optimizer.name]['results'] = d        

    solutions[objective.name] = estimators       

df = pd.DataFrame(results)
df_filename = "Benchmark Optimizations.csv"
save_df(df, figures, df_filename)

# --------------------------------------------------------------------------  #
# Render plots
for title, solution in solutions.items():        
    filename = title + " Optimization.html"
    filepath = os.path.join(figures, title)
    animate_optimization(estimators=solution, filepath=filepath)

#%#%%